from sqlalchemy import engine_from_config
from sqlalchemy.ext.declarative import declarative_base
from penelope.core.models.dbsession import DBSession

Base = declarative_base()


class classproperty(object):
    def __init__(self, getter):
        self.getter = getter

    def __get__(self, instance, owner):
        if instance:
            return self.getter(instance)
        else:
            return self.getter(owner)


from penelope.core.models.dashboard import Project, User, GlobalConfig, Role, OpenId,\
                                 PasswordResetToken, Application, CustomerRequest,\
                                 Estimation, Group, SavedQuery, Customer,\
                                 Contract, KanbanBoard

Project; User; GlobalConfig; PasswordResetToken; Role; SavedQuery
Application; Customer; CustomerRequest; Group; OpenId; Estimation; Contract
KanbanBoard

from penelope.core.models.tp import TimeEntry; TimeEntry
from penelope.trac import events; events


def includeme(config):
    if not config.registry.settings.get('test',False):
        initialize_sql(config)


def initialize_sql(config=None, engine=None):
    if not engine:
        engine = engine_from_config(config.registry.settings, 'sa.dashboard.')
    DBSession.configure(bind=engine)
    Base.metadata.bind = engine



# -*- coding: utf-8 -*-
import re
from trac.core import TracError
from trac.ticket.query import Query
from trac.ticket.model import Ticket
from trac.env import Environment
from pyramid.threadlocal import get_current_registry, get_current_request
from zope.interface import implements
from penelope.core.models.interfaces import ITicketStore
from penelope.core.models import DBSession


class TicketStore(object):
    implements(ITicketStore)

    def get_ticket(self, project_id, ticket_id):
        from penelope.core.models.dashboard import Project
        project = DBSession().query(Project).get(project_id)
        if project:
            settings = get_current_registry().settings
            tracenvs = settings.get('penelope.trac.envs')
            for trac in project.tracs:
                tracenv = Environment('%s/%s' % (tracenvs, trac.trac_name))
                tracenv.abs_href.base = trac.api_uri
                try:
                    t = Ticket(tracenv, ticket_id)
                    return (t.id, t.time_created, t.time_changed, t.values)
                except TracError:
                    return None

    def get_raw_ticket(self, project_id, ticket_id):
        from penelope.core.models.dashboard import Project
        project = DBSession().query(Project).get(project_id)
        if project:
            settings = get_current_registry().settings
            tracenvs = settings.get('penelope.trac.envs')
            for trac in project.tracs:
                tracenv = Environment('%s/%s' % (tracenvs, trac.trac_name))
                tracenv.abs_href.base = trac.api_uri
                try:
                    return Ticket(tracenv, ticket_id)
                except TracError:
                    return None

    def get_tickets_for_project(self, project, query=None, limit=None, not_invoiced=False):

        def queryWithDetails(env, qstr='status!=closed'):
            """
            Perform a ticket query, returning a list of ticket dictionaries.
            All queries will use stored settings for maximum number of results per
            page and paging options. Use `max=n` to define number of results to
            receive, and use `page=n` to page through larger result sets. Using
            `max=0` will turn off paging and return all results.
            """
            # TODO: gestione custom id^= (workaround)
            if "id^=" in qstr:
                startswith = re.search('id\^=([\d]*)', qstr).group(1)
                qstr = re.sub('id\^=[\d]*', '', qstr)
                qstr = re.sub('^&|&$|&&', '', qstr)
                if "max=" in qstr:
                    limit = re.search('max=([\d]*)', qstr).group(1)
                    qstr = re.sub('max=[\d]*', 'max=0', qstr)
                else:
                    limit = None
            else:
                startswith = None
                limit = None

            q = Query.from_string(env, qstr)
            out = []
            for t in q.execute():
                tid = t['id']
                if not startswith or str(tid).startswith(startswith):
                    out.append(t)
                    if limit and len(out) >= limit:
                        break

            # fill 'resolution' value (it is not provided by the above query)
            db = env.get_db_cnx()
            cursor = db.cursor()
            cursor.execute("SELECT id, resolution FROM ticket WHERE status='closed'")
            
            resolution = {}
            for row in cursor:
                resolution[row[0]] = row[1]

            # fill 'sensitive' value (TODO optimize, as it now retrieves all tickets)
            db = env.get_db_cnx()
            cursor = db.cursor()
            cursor.execute("SELECT ticket, value FROM ticket_custom WHERE name='sensitive' AND value='1'")
            sensitive = set()
            for row in cursor:
                sensitive.add(row[0])

            # fill 'customerrequest' value (TODO optimize, as it now retrieves all tickets)
            cursor = db.cursor()
            cursor.execute("SELECT t.id AS ticket, c.value FROM ticket t INNER JOIN ticket_custom c ON (t.id = c.ticket AND c.name = 'customerrequest') order by ticket")
            cr= {}
            for row in cursor:
                cr[row[0]]=row[1]

            for t in out:
                t['sensitive'] = int(t['id'] in sensitive)
                t['resolution'] = resolution.get(t['id'], '')
                t['cr'] = cr.get(t['id'], '')

            return out

        tickets = []
        settings = get_current_registry().settings
        tracenvs = settings.get('penelope.trac.envs')

        for trac in project.tracs:
            if not query:
                query = []
            if limit:
                query.append('max=%s' % limit)
            else:
                query.append('max=0')

            tracenv = Environment('%s/%s' % (tracenvs, trac.trac_name))
            tracenv.abs_href.base = trac.api_uri
            tickets.extend(queryWithDetails(tracenv, '&'.join(query)))

        if not_invoiced:
            cr_ids = [cr.id for cr in project.customer_requests if cr.workflow_state in ['created', 'estimated']]
            tickets = [t for t in tickets if t['cr'] in cr_ids]

        return tickets

    def get_tickets_for_request(self, customer_request, limit=None):
        cr = customer_request
        return self.get_tickets_for_project(project=cr.project,
                                            query=['customerrequest=%s' % (cr.id)],
                                            limit=limit)

    def get_number_of_tickets_per_cr(self, project):
        settings = get_current_registry().settings
        tracenvs = settings.get('penelope.trac.envs')

        for trac in project.tracs:
            env = Environment('%s/%s' % (tracenvs, trac.trac_name))
            db = env.get_db_cnx()
            cursor = db.cursor()
            cursor.execute("""SELECT c.value as cr, count(t.id) AS number FROM ticket t INNER JOIN ticket_custom c ON (t.id = c.ticket AND c.name = 'customerrequest') group by cr;""")
            tickets = cursor.fetchall()
            db.rollback()
        return dict(tickets)

    def get_all_ticket_crs(self, project):
        """
        returns a mapping of {ticket.id: cr_id}
        """
        def queryAllCustomerRequests(env):
            db = env.get_db_cnx()
            cursor = db.cursor()
            cursor.execute("SELECT ticket, value FROM ticket_custom WHERE name='customerrequest'")
            rows = cursor.fetchall() or []
            db.rollback()
            return rows

        ret = {}
        settings = get_current_registry().settings
        tracenvs = settings.get('penelope.trac.envs')
        for trac in project.tracs:
            tracenv = Environment('%s/%s' % (tracenvs, trac.trac_name))
            tracenv.abs_href.base = trac.api_uri
            for ticket_id, cr_id in queryAllCustomerRequests(tracenv):
                ret[ticket_id] = cr_id

        return ret

    def get_requests_from_tickets(self, project, ticket_ids):

        def queryCustomerRequestsByTicktes(env, ticket_ids):
            db = env.get_db_cnx()
            cursor = db.cursor()
            cursor.execute("""SELECT ticket, value FROM ticket_custom
                WHERE name='customerrequest' AND ticket IN %(tickets)s;""",
                    {'tickets': tuple(ticket_ids)})
            customer_request_ids = cursor.fetchall() or []
            db.rollback()
            return customer_request_ids

        ticket_cr = []
        settings = get_current_registry().settings
        tracenvs = settings.get('penelope.trac.envs')
        for trac in project.tracs:
            tracenv = Environment('%s/%s' % (tracenvs, trac.trac_name))
            tracenv.abs_href.base = trac.api_uri
            ticket_cr.extend(queryCustomerRequestsByTicktes(tracenv, ticket_ids))
        return ticket_cr

    def add_tickets(self, project, customerrequest, tickets, reporter, notify=False):
        from trac.ticket.notification import TicketNotifyEmail
        from trac.util.text import exception_to_unicode
        from penelope.core.models.dashboard import User

        settings = get_current_registry().settings
        tracenvs = settings.get('penelope.trac.envs')
        request = get_current_request()

        for trac in project.tracs:
            for t in tickets:
                owner = DBSession.query(User).get(t['owner'])
                ticket = {'summary': t['summary'],
                        'description': t['description'],
                        'customerrequest': customerrequest.id,
                        'reporter': reporter.email,
                        'type': 'task',
                        'priority': 'major',
                        'milestone': 'Backlog',
                        'owner': owner.email,
                        'status': 'new'}
                tracenv = Environment('%s/%s' % (tracenvs, trac.trac_name))
                tracenv.abs_href.base = trac.api_uri
                t = Ticket(tracenv)
                t.populate(ticket)
                t.insert()
                if notify:
                    try:
                        tn = TicketNotifyEmail(tracenv)
                        tn.notify(t, newticket=True)
                    except Exception, e:
                        request.add_message('Failure sending notification on creation '
                        'of a ticket #%s: %s' % (t.id, exception_to_unicode(e)), 'error')

ticket_store = TicketStore()



# -*- coding: utf-8 -*-

import cgi
import colander
import deform

from colander import SchemaNode
from pyramid.view import view_config
from pyramid.response import Response

from penelope.core.models import DBSession, SavedQuery
from penelope.core.lib.widgets import DeleteButton, FavoriteButton, RenameButton, PorInlineForm


class SavedQuerySchema(colander.MappingSchema):
    sq_id = SchemaNode(colander.Int(),
                       widget=deform.widget.HiddenWidget(),
                       title=u'')

    report_name = SchemaNode(colander.String(),
                             widget=deform.widget.HiddenWidget(),
                             title=u'')

    query_name = SchemaNode(colander.String(),
                            title=u'Query name')



def render_saved_query_form(request):
    current_uid = request.authenticated_user.id
    qry = DBSession.query(SavedQuery)

    # retrieve a saved query by the same user...
    qry = qry.filter(SavedQuery.author_id==current_uid)
    # ..for the current report..
    qry = qry.filter(SavedQuery.report_name==request.view_name)
    # ..with the current parameters
    qry = qry.filter(SavedQuery.query_string=='?'+request.query_string)

    # XXX what if there are two queries with the exact same parameters?

    savedqry = qry.first()
    report_name = request.view_name

    if savedqry:
        appstruct = {
                'sq_id': savedqry.id,
                'query_name': savedqry.query_name,
                'report_name': report_name,
                }
        buttons = [
                RenameButton(title=u'Rename', type='button', name='submit_edit'),
                DeleteButton(title=u'Delete', type='button', name='submit_delete'),
                ]
    else:
        appstruct = {
                'report_name': report_name,
                }
        buttons = [
                FavoriteButton(title=u'Save', type='button', name="submit_add"),
                ]


    form = PorInlineForm(SavedQuerySchema().clone(),
                         formid='saved_query_form',
                         action='save_query',
                         method='POST',
                         buttons=buttons,
                         )
    return form.render(appstruct=appstruct)



@view_config(name='save_query', route_name='reports', permission='reports_my_entries')
def save_query(context, request):
    current_uid = request.authenticated_user.id

    query_meta = cgi.parse_qs(request.POST['query_meta'])

    if not 'query_name' in query_meta:
        return Response(u"Please specify a query name.", status=409)

    query_name = query_meta['query_name'][0]

    taken = DBSession.query(SavedQuery).filter(SavedQuery.author_id==current_uid).filter(SavedQuery.query_name==query_name).count()

    submit_type = request.POST['submit_type']

    if submit_type == 'submit_edit':
        if taken:
            return Response(u"Name already in use: '%s'." % query_name, status=409)
        sq_id = query_meta['sq_id'][0]
        qry = DBSession.query(SavedQuery)
        qry = qry.filter(SavedQuery.author_id==current_uid)
        qry = qry.filter(SavedQuery.id==sq_id)
        sq = qry.one()
        sq.query_name = query_name
        return Response(u"The query has been renamed as '%s'." % query_name)
    elif submit_type == 'submit_delete':
        sq_id = query_meta['sq_id'][0]
        qry = DBSession.query(SavedQuery)
        qry = qry.filter(SavedQuery.author_id==current_uid)
        qry = qry.filter(SavedQuery.id==sq_id)
        sq = qry.one()
        DBSession.delete(sq)
        return Response(u"The saved query has been deleted.")
    elif submit_type == 'submit_add':
        if taken:
            return Response(u"Name already in use: '%s'." % query_name, status=409)
        # add
        sq = SavedQuery(query_name=query_name,
                        report_name=query_meta['report_name'][0],
                        query_string=request.POST['query_string'],
                        author_id = current_uid)
        DBSession.add(sq)
        return Response(u"The query has been saved as '%s'." % query_name)



# -*- coding: utf-8 -*-

import colander
import collections
import logging
import sqlalchemy as sa

from deform import ValidationFailure
from pyramid.renderers import render
from pyramid.view import view_config
from repoze.workflow import get_workflow, WorkflowError
from sqlalchemy.orm import lazyload
from webhelpers.html.builder import HTML

from penelope.core import fanstatic_resources
from penelope.core.lib.helpers import ticket_url, unicodelower
from penelope.core.lib.widgets import SearchButton, PorInlineForm
from penelope.core.models import DBSession, Project, TimeEntry, User, CustomerRequest, Contract, Customer
from penelope.core.models.tickets import ticket_store
from penelope.core.reports import fields
from penelope.core.reports.favourites import render_saved_query_form
from penelope.core.reports.queries import qry_active_projects, te_filter_by_customer_requests, filter_users_with_timeentries, te_filter_by_contracts
from penelope.core.reports.validators import validate_period

log = logging.getLogger(__name__)


class StateChangeReport(object):

    def __init__(self, context, request):
        self.context = context
        self.request = request

    class StateChangeSchema(colander.MappingSchema):
        customer_id = fields.customer_id.clone()

        project_id = fields.project_id.clone()
        date_from = fields.date_from.clone()
        date_to = fields.date_to.clone()
        users = fields.users.clone()
        contracts = fields.contracts.clone()
        customer_requests = fields.customer_requests.clone()
        workflow_states = fields.workflow_states.clone()
        invoice_number = fields.invoice_number.clone()

    def search(self, customer_id, project_id, date_from, date_to,
               users, customer_requests, invoice_number, workflow_states, contracts):

        # also search archived projects, if none are specified
        qry = DBSession.query(TimeEntry).join(TimeEntry.project).join(Project.customer).outerjoin(TimeEntry.author)
        qry = qry.options(lazyload(TimeEntry.project, TimeEntry.author, Project.customer))

        if customer_id is not colander.null:
            qry = qry.filter(Project.customer_id == customer_id)

        if project_id is not colander.null:
            qry = qry.filter(TimeEntry.project_id == project_id)

        if date_from is not colander.null:
            qry = qry.filter(TimeEntry.date>=date_from)

        if date_to is not colander.null:
            qry = qry.filter(TimeEntry.date<=date_to)

        if users:
            qry = qry.filter(TimeEntry.author_id.in_(users))

        if invoice_number:
            qry = qry.filter(TimeEntry.invoice_number==invoice_number)

        if workflow_states:
            qry = qry.filter(TimeEntry.workflow_state.in_(workflow_states))

        qry = qry.filter(te_filter_by_customer_requests(customer_requests, request=self.request))
        qry = qry.filter(te_filter_by_contracts(contracts))
        qry = qry.order_by(sa.desc(TimeEntry.date), sa.desc(TimeEntry.start), sa.desc(TimeEntry.creation_date))

        time_entries = self.request.filter_viewables(qry)

        proj_tickets = collections.defaultdict(set)
        for te in time_entries:
            if te.ticket is None:
                continue
            proj_tickets[te.project_id].add(te.ticket)

        projectsmap = {}
        for project_id, ticket_ids in proj_tickets.items():
            project = DBSession.query(Project).get(project_id)
            projectsmap[project_id] = dict(ticket_store.get_requests_from_tickets(
                                            project, tuple(ticket_ids)))

        tkts = {}
        for project_id, ticket_ids in proj_tickets.items():
            project = DBSession.query(Project).get(project_id)
            for tkt in ticket_store.get_tickets_for_project(project):
                tkts[(project_id, tkt['id'])] = tkt

        entries_tree = collections.defaultdict(lambda: collections.defaultdict(list))

        for te in time_entries:
            customer_request = te.customer_request
            entry = {
                        'customer': te.project.customer.name.strip(),
                        'project': te.project.name.strip(),
                        'user': te.author.fullname.strip(),
                        'date': te.date.strftime('%Y-%m-%d'),
                        'contract': te.customer_request.contract,
                        'description': te.description,
                        'hours_str': te.hours_str,
                        'workflow_state': te.workflow_state,
                        'tickettype': te.tickettype,
                        'invoice_number': te.invoice_number,
                        'location': te.location,
                        'id': te.id,
                    }

            if te.ticket is None:
                ticket_description = '(no ticket)'
            else:
                tkt = tkts[(te.project_id, te.ticket)]
                href = ticket_url(request=self.request,
                                  project=te.project,
                                  ticket_id=te.ticket)
                ticket_description = HTML.SPAN(
                                            HTML.A('#%(id)s - %(summary)s' % tkt, href=href),
                                            ' ',
                                            HTML.SUP(tkt['type']),
                                            )

            entries_tree[customer_request][ticket_description].append(entry)

        return {
                'entries_tree': entries_tree,
                }

    def state_cr_change(self):
        new_state = self.request.POST['new_state']
        new_cr = self.request.POST['new_cr']
        invoice_number = self.request.POST['invoice_number']

        te_ids = set(int(s[3:])
                     for s, checkbox_state in self.request.POST.iteritems()
                     if s.startswith('te_') and checkbox_state=='on')

        qry = DBSession.query(TimeEntry).filter(TimeEntry.id.in_(te_ids))

        done_state = set()
        done_cr = set()
        errors = {}

        for te in qry:
            if new_state:
                try:
                    workflow = get_workflow(te, te.__class__.__name__)
                    workflow.transition_to_state(te, self.request, new_state, skip_same=True)
                    done_state.add(te.id)
                    if new_state == 'invoiced' and invoice_number:
                        te.invoice_number = invoice_number
                except WorkflowError as msg:
                    errors[te.id] = msg
            if new_cr:
                done_cr.add(te.id)
                te.customer_request_id = new_cr

        return done_state, done_cr, errors


    @view_config(name='report_state_change', route_name='reports', renderer='skin', permission='reports_state_change')
    def __call__(self):

        done_state = set()
        done_cr = set()
        errors = {}
        if self.request.POST:
            done_state, done_cr, errors = self.state_cr_change()
            if done_state:
                self.request.add_message('State changed for %d time entries.' % len(done_state))
            if done_cr:
                self.request.add_message('Customer request changed for %d time entries.' % len(done_cr))

        # GET parameters for the search form

        fanstatic_resources.report_te_state_change.need()
        schema = self.StateChangeSchema(validator=validate_period).clone()
        projects = self.request.filter_viewables(qry_active_projects())

        # select customers that have some active project
        customers = sorted(set(p.customer for p in projects), key=unicodelower)

        users = DBSession.query(User).order_by(User.fullname)
        users = filter_users_with_timeentries(users)
        customer_requests = DBSession.query(CustomerRequest).order_by(CustomerRequest.name)
        contracts = DBSession.query(Contract).order_by(Contract.name)

        form = PorInlineForm(schema,
                             formid='te_state_change',
                             method='GET',
                             buttons=[
                                 SearchButton(title=u'Search'),
                             ])

        workflow = get_workflow(TimeEntry(), 'TimeEntry')

        all_wf_states = [
                            (state, workflow._state_data[state]['title'] or state)
                            for state in workflow._state_order
                        ]

        form['workflow_states'].widget.values = all_wf_states

        form['customer_id'].widget.values = [('', '')] + [(str(c.id), c.name) for c in customers]
        # don't validate as it might be an archived customer
        form['project_id'].widget.values = [('', '')] + [(str(p.id), p.name) for p in projects]
        # don't validate as it might be an archived project

        form['users'].widget.values = [(str(u.id), u.fullname) for u in users]
        form['customer_requests'].widget.values = [(str(c.id), c.name) for c in customer_requests]
        form['contracts'].widget.values = [(str(c.id), c.name) for c in contracts]

        controls = self.request.GET.items()

        if not controls:
            # the form is empty
            return {
                    'form': form.render(),
                    'saved_query_form': render_saved_query_form(self.request),
                    'qs':'',
                    'result_table': '',
                    }

        try:
            appstruct = form.validate(controls)
        except ValidationFailure as e:
            return {
                    'form': e.render(),
                    'saved_query_form': render_saved_query_form(self.request),
                    'qs':'',
                    'result_table': '',
                    }

        entries_detail = self.search(**appstruct)

        if appstruct.get('project_id'):
            contracts = DBSession().query(Project).get(appstruct['project_id']).contracts
        else:
            contracts = [item for sublist in 
                                [p.contracts for p in 
                                DBSession().query(Customer).get(appstruct['customer_id']).projects]
                            for item in sublist]

        result_table = render('penelope.core:reports/templates/state_change.pt',
                              {
                                  'entries_tree': entries_detail['entries_tree'],
                                  'all_wf_states': all_wf_states,
                                  'contracts': contracts,
                                  'wf_state_names': dict((ws[0], ws[1]) for ws in all_wf_states),
                                  'done_state': done_state,
                                  'done_cr': done_cr,
                                  'errors': errors,
                              },
                              request=self.request)

        return {
                'form': form.render(appstruct=appstruct),
                'saved_query_form': render_saved_query_form(self.request),
                'qs': self.request.query_string,
                'result_table': result_table,
                }


class UpdateTimeEntries(object):

    def __init__(self, context, request):
        self.context = context
        self.request = request

    @view_config(name='update_te_cr', route_name='reports', renderer='skin', permission='reports_state_change')
    def __call__(self):
        te_ids = set(int(s[3:])
                     for s, checkbox_state in self.request.POST.iteritems()
                     if s.startswith('te_') and checkbox_state=='on')
        tes = DBSession.query(TimeEntry).filter(TimeEntry.id.in_(te_ids))
        new_state = self.request.POST.get('new_state', None)
        new_cr = self.request.POST.get('new_cr', None)
        if new_cr:
            new_cr = DBSession.query(CustomerRequest).get(new_cr)

        return {'back_url': '%s/reports/report_state_change?%s' % (self.request.application_url, self.request.query_string),
                'post': self.request.POST.items(),
                'tes': tes,
                'new_state': new_state,
                'new_cr': new_cr}



# -*- coding: utf-8 -*-

from copy import deepcopy
import json

from pyramid.httpexceptions import HTTPFound
from pyramid import httpexceptions as exc
from pyramid_skins import SkinObject
from repoze.workflow import get_workflow, WorkflowError

from fa.bootstrap import actions

from penelope.core.backlog import Backlog
from penelope.core.models import dashboard
from penelope.core.models import DBSession
from penelope.core.forms import ModelView


def configurate(config):
    #custom view for project
    config.formalchemy_model_view('admin',
        request_method='GET',
        permission='view',
        name='',
        attr='documentation',
        renderer='fa.bootstrap:templates/admin/show.pt',
        model='penelope.core.models.dashboard.Project',
        view=ProjectModelView)

    config.formalchemy_model_view('admin',
        request_method='POST',
        permission='delete',
        name='delete',
        attr='delete',
        renderer='fa.bootstrap:templates/admin/edit.pt',
        model='penelope.core.models.dashboard.Project',
        view=ProjectModelView)

    config.formalchemy_model_view('admin',
        renderer='penelope.core.forms:templates/project_listing.pt',
        attr='datatable',
        context='pyramid_formalchemy.resources.ModelListing',
        request_method='GET',
        permission='listing',
        model='penelope.core.models.dashboard.Project',
        view=ProjectModelView)

    #custom view for adding an application to the project
    config.formalchemy_model_view('admin',
        request_method='GET',
        permission='new',
        name='add_group',
        attr='add_group',
        renderer='penelope.core.forms:templates/new.pt',
        model='penelope.core.models.dashboard.Project',
        view=ProjectModelView)

    config.formalchemy_model_view('admin',
        request_method='POST',
        permission='new',
        name='add_group',
        attr='add_group',
        renderer='penelope.core.forms:templates/new.pt',
        model='penelope.core.models.dashboard.Project',
        view=ProjectModelView)

    #custom view for adding an application to the project
    config.formalchemy_model_view('admin',
        request_method='GET',
        permission='new',
        name='add_application',
        attr='add_application',
        renderer='penelope.core.forms:templates/new.pt',
        model='penelope.core.models.dashboard.Project',
        view=ProjectModelView)

    config.formalchemy_model_view('admin',
        request_method='POST',
        permission='new',
        name='add_application',
        attr='add_application',
        renderer='penelope.core.forms:templates/new.pt',
        model='penelope.core.models.dashboard.Project',
        view=ProjectModelView)

    #custom view for configuration section
    config.formalchemy_model_view('admin',
        request_method='GET',
        permission='edit',
        name='configuration',
        attr='configuration',
        renderer='penelope.core.forms:templates/configuration.pt',
        model='penelope.core.models.dashboard.Project',
        view=ProjectModelView)

    #custom view for documentation section
    config.formalchemy_model_view('admin',
        request_method='GET',
        permission='metadata',
        name='metadata',
        attr='show',
        renderer='penelope.core.forms:templates/project.pt',
        model='penelope.core.models.dashboard.Project',
        view=ProjectModelView)

    #custom view for customer_requests section
    config.formalchemy_model_view('admin',
        request_method='GET',
        permission='view',
        name='customer_requests',
        attr='customer_requests',
        model='penelope.core.models.dashboard.Project',
        view=ProjectModelView)

    #custom view for customer_requests section
    config.formalchemy_model_view('admin',
        request_method='POST',
        permission='edit',
        name='customer_requests',
        attr='update_cr_states',
        model='penelope.core.models.dashboard.Project',
        view=ProjectModelView)

    #custom view for contracts section
    config.formalchemy_model_view('admin',
        request_method='GET',
        permission='view',
        name='contracts',
        attr='contracts',
        renderer='penelope.core.forms:templates/contracts.pt',
        model='penelope.core.models.dashboard.Project',
        view=ProjectModelView)


    #custom view for tickets section
    config.formalchemy_model_view('admin',
        request_method='GET',
        permission='view',
        name='tickets',
        attr='tickets',
        renderer='penelope.core.forms:templates/tickets.pt',
        model='penelope.core.models.dashboard.Project',
        view=ProjectModelView)

    #custom view for time_entries section
    config.formalchemy_model_view('admin',
        request_method='GET',
        permission='time_entries',
        name='time_entries',
        attr='time_entries',
        renderer='penelope.core.forms:templates/time_entries.pt',
        model='penelope.core.models.dashboard.Project',
        view=ProjectModelView)

    #custom view for documentation section
    config.formalchemy_model_view('admin',
        request_method='GET',
        permission='view',
        name='applications',
        attr='applications',
        renderer='penelope.core.forms:templates/configurate_apps.pt',
        model='penelope.core.models.dashboard.Project',
        view=ProjectModelView)

    #custom view for adding a customer request to the project
    config.formalchemy_model_view('admin',
        request_method='GET',
        permission='new',
        name='add_customer_request',
        attr='add_customer_request',
        renderer='penelope.core.forms:templates/new.pt',
        model='penelope.core.models.dashboard.Project',
        view=ProjectModelView)

    config.formalchemy_model_view('admin',
        request_method='POST',
        permission='new',
        name='add_customer_request',
        attr='add_customer_request',
        renderer='penelope.core.forms:templates/new.pt',
        model='penelope.core.models.dashboard.Project',
        view=ProjectModelView)

    #custom view for adding a contract to the project
    config.formalchemy_model_view('admin',
        request_method='GET',
        permission='new',
        name='add_contract',
        attr='add_contract',
        renderer='penelope.core.forms:templates/new.pt',
        model='penelope.core.models.dashboard.Project',
        view=ProjectModelView)

    config.formalchemy_model_view('admin',
        request_method='POST',
        permission='new',
        name='add_contract',
        attr='add_contract',
        renderer='penelope.core.forms:templates/new.pt',
        model='penelope.core.models.dashboard.Project',
        view=ProjectModelView)


    config.formalchemy_model_view('admin',
        request_method='GET',
        permission='view',
        name='toggle_favorite',
        attr='toggle_favorite',
        model='penelope.core.models.dashboard.Project',
        view=ProjectModelView)


class ProjectModelView(ModelView):
    actions_categories = ('buttons', 'tabs', 'subtabs')
    defaults_actions = deepcopy(actions.defaults_actions)

    def __init__(self, *args, **kwargs):
        super(ProjectModelView, self).__init__(*args, **kwargs)
        self.defaults_actions.update(show_buttons=actions.Actions(actions.edit))
        self.defaults_actions.update(listing_buttons=actions.Actions())
        self.defaults_actions.update(documentation_buttons=actions.Actions())

    def delete(self):
        project = self.context.get_instance()
        request = self.request
        STOP = False
        if len(project.groups):
            request.add_message(u'Project has existing groups. Please remove them first.', type='danger')
            STOP = True
        if len(project.applications):
            request.add_message(u'Project has existing applications. Please remove them first.', type='danger')
            STOP = True
        if len(project.customer_requests):
            request.add_message(u'Project has existing customer_requests. Please remove them first.', type='danger')
            STOP = True
        if len(project.time_entries):
            request.add_message(u'Project has existing time entries. Please remove them first.', type='danger')
            STOP = True

        if STOP:
            raise exc.HTTPFound(location=request.fa_url('Project', project.id))
        else:
            request.fa_parent_url = lambda : request.fa_url(request.model_name)
            return self.force_delete()

    def add_group(self, *args, **kwargs):
        self.request.model_class = dashboard.Group
        self.request.model_name = dashboard.Group.__name__
        self.request.form_action = ['Project', self.request.model_id, 'add_group']
        return self.new()

    def add_application(self, *args, **kwargs):
        self.request.model_class = dashboard.Application
        self.request.model_name = dashboard.Application.__name__
        self.request.form_action = ['Project', self.request.model_id, 'add_application']
        return self.new()

    def add_customer_request(self, *args, **kwargs):
        self.request.model_class = dashboard.CustomerRequest
        self.request.model_name = dashboard.CustomerRequest.__name__
        self.request.form_action = ['Project', self.request.model_id, 'add_customer_request']
        return self.new()

    def add_contract(self, *args, **kwargs):
        self.request.model_class = dashboard.Contract
        self.request.model_name = dashboard.Contract.__name__
        self.request.form_action = ['Project', self.request.model_id, 'add_contract']
        return self.new()

    def all_customer_requests(self):
        session = DBSession()
        return self.request.filter_viewables(
                        session.query(dashboard.CustomerRequest)\
                               .filter(dashboard.CustomerRequest.project ==\
                                       self.context.get_instance()))

    @actions.action()
    def time_entries(self, *args, **kwargs):
        context = self.context.get_instance()
        page = self.get_page(collection=context.time_entries)
        pager = page.pager(**self.pager_args)
        return self.render(pager=pager, items=page)

    @actions.action()
    def tickets(self, *args, **kwargs):
        return self.render(customer_requests=self.all_customer_requests())

    @actions.action()
    def contracts(self, *args, **kwargs):
        context = self.context.get_instance()
        page = self.get_page(collection=context.contracts)
        pager = page.pager(**self.pager_args)
        return self.render(pager=pager, items=page)

    @actions.action()
    def customer_requests(self, *args, **kwargs):
        project = self.context.get_instance()
        params = Backlog(self.request).backlog(projects=[project])
        params['request'] = self.request
        params['context'] = self.request.context
        params['all_contracts'] = project.contracts_by_state()
        params['multiple_bgb'] = False
        params['number_of_tickets_per_cr'] = project.get_number_of_tickets_per_cr()
        return SkinObject('tekken')(**params)

    def update_cr_states(self):
        context = self.context.get_instance()
        n_cr = 0

        cr_ids = self.request.POST.getall('cr')
        new_state = self.request.POST.get('new_state')
        for cr in DBSession().query(dashboard.CustomerRequest).filter(dashboard.CustomerRequest.id.in_(cr_ids)):
            try:
                workflow = get_workflow(cr, cr.__class__.__name__)
                workflow.transition_to_state(cr, self.request, new_state, skip_same=True)
                n_cr += 1
            except WorkflowError as msg:
                self.request.add_message(msg, 'warning')

        self.request.add_message('%d customer requets updated.' % n_cr)
        raise exc.HTTPFound(location=self.request.fa_url('Project', context.id, 'customer_requests'))

    @actions.action('documentation')
    def documentation(self, *args, **kwargs):
        raise HTTPFound('%s/applications' % self.request.current_route_url())

    @actions.action('configuration')
    def configuration(self, *args, **kwargs):
        context = self.context.get_instance()
        return self.render(groups=context.groups)

    @actions.action('applications')
    def applications(self, *args, **kwargs):
        context = self.context.get_instance()
        return self.render(applications=context.applications)

    def toggle_favorite(self, *args, **kwargs):
        user = self.request.authenticated_user
        context = self.context.get_instance()
        if user not in context.favorite_users:
            context.favorite_users.append(user)
            self.request.add_message(u'Project %s has been marked as favorite.' % context.name, 'success')
        else:
            context.favorite_users.remove(user)
            self.request.add_message(u'Project %s has been unmarked as favorite.' % context.name, 'success')
        raise HTTPFound(self.request.fa_url('Project', context.id))


    @actions.action('listing')
    def datatable(self, **kwargs):
        result = super(ProjectModelView, self).datatable(**kwargs)

        fs = result['fs']
        fs.configure(pk=True, readonly=True)
        viewed_projects = set(fs.rows.items)

        columns = ['name', 'customer', 'manager', 'customer_requests', 'groups', 'activated', 'id']
        self.pick_columns(fs, columns)

        user = self.request.authenticated_user
        favorite_projects = set(p.id for p in set(user.favorite_projects).intersection(viewed_projects))

        return dict(result,
                    columns=columns,
                    js_favorite_projects=json.dumps(list(favorite_projects)))



import os
import argparse
import sys
import beaker
import csv
import tempfile

import gdata.data
import gdata.docs.client
import gdata.docs.data
import gdata.docs.service
import gdata.sample_util


from pyramid.paster import get_appsettings, setup_logging
from datetime import datetime, date, timedelta
from sqlalchemy.orm import mapper
from sqlalchemy import engine_from_config, extract
from sqlalchemy import MetaData, Table, and_, or_
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import relationship, backref

from penelope.core.models.dashboard import Project, CustomerRequest, Trac, Estimation
from penelope.core.models.tp import TimeEntry, timedelta_as_work_days
from penelope.core.models.dbsession import DBSession
from penelope.core.models import Base


beaker.cache.cache_regions.update(dict(calculate_matrix={}))


def create_client():
    client = gdata.docs.client.DocsClient(source='GDataDocumentsListAPISample-v1.0')
    try:
        gdata.sample_util.authorize_client(
            client,
            1,
            service=client.auth_service,
            source=client.source,
            scopes=client.auth_scopes
        )
    except gdata.client.BadAuthentication:
        exit('Invalid user credentials given.')
    except gdata.client.Error:
        exit('Login Error')
    return client


def upload_file(client, path, title, collection):

    def get_upload_folder():
        folders = client.get_doclist(uri='/feeds/default/private/full?title=%s&category=folder' % collection)
        for folder in folders.entry:
            return folder

    content_type = gdata.docs.service.SUPPORTED_FILETYPES['CSV']
    entry = client.Upload(path, title=title, content_type=content_type)
    folder = get_upload_folder()
    if folder:
        client.move(entry, folder)
    print 'Uploaded...'


def tickets_for_cr(metadata, session, trac_name, cr_id=None):

    class Ticket(object):
        @hybrid_property
        def date(self):
            return datetime.fromtimestamp(self.time/1000000)

        def last_history(self, name, year):
            from_year = [a for a in self.history \
                                   if a.date.year == year and a.field == name]
            from_year.sort(key=lambda a: a.date, reverse=True)
            for history in from_year:
                return history.newvalue
            return getattr(self, name)

        @property
        def close_date(self):
            status = [a for a in self.history if a.field == 'status']
            status.sort(key=lambda a: a.date, reverse=True)
            for history in status:
                if history.newvalue == 'closed':
                    return history.date
            return None


    class TicketChange(object):
        @hybrid_property
        def date(self):
            return datetime.fromtimestamp(self.time/1000000)


    class TicketCustom(object):
        def __init__(self, ticket, name, value):
            self.ticket = ticket
            self.name = name
            self.value = value

        @property
        def unicode_value(self):
            return self.value.encode('utf8','ignore')

    ticket = Table('ticket', metadata, autoload=True, schema='trac_%s' % trac_name)
    ticket_custom = Table('ticket_custom', metadata, autoload=True, schema='trac_%s' % trac_name)
    ticket_change = Table('ticket_change', metadata, autoload=True, schema='trac_%s' % trac_name)
    mapper(TicketCustom, ticket_custom)
    mapper(TicketChange, ticket_change, properties={
            'realticket':relationship(Ticket, primaryjoin=ticket_change.c.ticket==ticket.c.id,
                                      foreign_keys=[ticket_change.c.ticket],
                                      backref=backref('history')),
    })
    mapper(Ticket, ticket, properties={
            'customer_request':relationship(TicketCustom, primaryjoin=and_(ticket_custom.c.ticket==ticket.c.id,
                                                                          ticket_custom.c.name=='customerrequest'),
                                           foreign_keys=[ticket.c.id]),
            'open_by_customer':relationship(TicketCustom, primaryjoin=and_(ticket_custom.c.ticket==ticket.c.id,
                                                                           ticket_custom.c.name=='esogeno'),
                                            foreign_keys=[ticket.c.id]),
            'issue_type':relationship(TicketCustom, primaryjoin=and_(ticket_custom.c.ticket==ticket.c.id,
                                                                     ticket_custom.c.name=='issuetype'),
                                      foreign_keys=[ticket.c.id]),
            })

    query = session.query(Ticket)
    if cr_id:
        query = query.outerjoin(TicketCustom, TicketCustom.ticket==Ticket.id)\
                     .filter(and_(TicketCustom.name=='customerrequest',
                                  TicketCustom.value==cr_id))
    return query, Ticket, TicketCustom


class Quality(argparse.Action):
    def __call__(self, parser, namespace, values, option_string):
        super(Quality, self).__init__(parser, namespace, values, option_string)
        self.namespace = namespace
        config_uri = namespace.configuration
        setup_logging(config_uri)
        settings = get_appsettings(config_uri, name='dashboard')
        engine = engine_from_config(settings, 'sa.dashboard.')
        DBSession.configure(bind=engine)
        Base.metadata.bind = engine
        Base.metadata.create_all()
        self.metadata = MetaData(engine)

        #if not namespace.google and not namespace.filename:
        #    raise argparse.ArgumentTypeError(u'You need to pass filename or google.')

        configuration = {}
        tmp = tempfile.NamedTemporaryFile(suffix='.csv')
        namespace.filename = tmp.name
        tmp.close()
        configuration['filename'] = namespace.filename
        print 'Creating file %s' % namespace.filename
        setattr(namespace, option_string.strip('--'), configuration)


class QualityProject(Quality):
    def __call__(self, parser, namespace, values, option_string):
        super(QualityProject, self).__call__(parser, namespace, values, option_string)
        session = DBSession()
        with open(namespace.filename, 'wb') as ofile:
            writer = csv.writer(ofile, dialect='excel')
            writer.writerow(['Project ID', 'Customer', 'Project creation year',
                             'Project creation month', 'Project creation day',
                             'Project completion year', 'Project completion month',
                             'Project completion day'])
            for project in session.query(Project.id, Project.completion_date,
                                   Project.customer_id, Project.creation_date)\
                                  .outerjoin(TimeEntry, TimeEntry.project_id==Project.id)\
                                  .filter(extract('year', TimeEntry.date) == namespace.year)\
                                  .distinct():

                writer.writerow([project.id, project.customer_id, 
                                 project.creation_date.strftime('%Y'),
                                 project.creation_date.strftime('%m'),
                                 project.creation_date.strftime('%d'),
                                 project.completion_date and project.completion_date.strftime('%Y') or '',
                                 project.completion_date and project.completion_date.strftime('%m') or '',
                                 project.completion_date and project.completion_date.strftime('%d') or '',
                                 ])


class QualityCR(Quality):
    def __call__(self, parser, namespace, values, option_string):
        super(QualityCR, self).__call__(parser, namespace, values, option_string)
        session = DBSession()
        with open(namespace.filename, 'wb') as ofile:
            writer = csv.writer(ofile, dialect='excel')
            writer.writerow(['CR ID', 'Description', 'Customer', 'CR state',
                             'Estimation in days', 'TE duration in days - current year',
                             'TE sistem/install in days - current year',
                             'TE duration in days - overall', 'TE sistem/install in days - overall'])
            crs = session.query(CustomerRequest.id,
                                CustomerRequest.name,
                                CustomerRequest.workflow_state,
                                CustomerRequest.project_id,
                                Project.customer_id).\
                   outerjoin(Project, CustomerRequest.project_id == Project.id).distinct()

            for cr in crs:

                all_entries = session.query(TimeEntry).filter_by(customer_request_id=cr.id)
                if not all_entries.count():
                    continue

                estimations = sum([a.days for a in \
                                session.query(Estimation.days)\
                                       .filter_by(customer_request_id=cr.id)])

                current_year_entries = session.query(TimeEntry)\
                                              .filter_by(customer_request_id=cr.id)\
                                              .filter(extract('year', TimeEntry.date) == namespace.year)

                total_hours = timedelta_as_work_days(sum([a.hours for a in all_entries], timedelta()))
                total_dev_hours = timedelta_as_work_days(sum([a.hours for a in
                                            all_entries.filter(or_(TimeEntry.description.ilike('%install%'),
                                                                    TimeEntry.description.ilike('%sistem%')))], timedelta()))
                current_year_hours = timedelta_as_work_days(sum([a.hours for a in current_year_entries], timedelta()))
                current_year_dev_hours = timedelta_as_work_days(sum([a.hours for a in
                                                    current_year_entries.filter(or_(TimeEntry.description.ilike('%install%'),
                                                                                    TimeEntry.description.ilike('%sistem%')))], timedelta()))

                writer.writerow([cr.id, cr.name.encode('utf8'), cr.customer_id,
                    cr.workflow_state, estimations, current_year_hours,
                    current_year_dev_hours, total_hours, total_dev_hours])


class QualityTicket(Quality):
    def __call__(self, parser, namespace, values, option_string):
        super(QualityTicket, self).__call__(parser, namespace, values, option_string)
        session = DBSession()
        with open(namespace.filename, 'wb') as ofile:
            writer = csv.writer(ofile, dialect='excel')
            writer.writerow(['Ticket ID', 'Customer',
                             'Ticket creation year',
                             'Ticket creation month',
                             'Ticket creation day',
                             'Ticket completion year',
                             'Ticket completion moneth',
                             'Ticket completion day',
                             'Ticket state', 'Ticket last owner', 'Ticket types',
                             'Ticket opened by customer', 'Problem nature'])

            for pr in session.query(Project.customer_id, Project.id, Trac.trac_name)\
                             .outerjoin(Trac, Project.id==Trac.project_id)\
                             .outerjoin(TimeEntry, TimeEntry.project_id==Project.id)\
                             .filter(TimeEntry.project_id==Project.id)\
                             .filter(extract('year', TimeEntry.date) == namespace.year).distinct():
                tickets, Ticket, TicketCustom = tickets_for_cr(self.metadata, session, pr.trac_name)
                tickets_in_year = session.query(Ticket)\
                                         .outerjoin(TimeEntry,
                                                    and_(TimeEntry.project_id==pr.id,
                                                         TimeEntry.ticket==Ticket.id))\
                                         .filter(extract('year', TimeEntry.date) == namespace.year).distinct()
                for ticket in tickets_in_year:
                    last_status = ticket.last_history('status', namespace.year)
                    close_date = ticket.close_date
                    all_types = set([])
                    for h in ticket.history:
                        if h.field == 'type':
                            all_types.update([h.oldvalue])
                    all_types.update([ticket.type])
                    all_types = '|'.join(all_types).encode('utf8','ignore')
                    writer.writerow(
                          [ticket.id, pr.customer_id,
                           ticket.date.strftime('%Y'),
                           ticket.date.strftime('%m'),
                           ticket.date.strftime('%d'),
                           close_date and close_date.strftime('%Y') or '',
                           close_date and close_date.strftime('%m') or '',
                           close_date and close_date.strftime('%d') or '',
                           last_status, ticket.owner, all_types,
                           ticket.open_by_customer and ticket.open_by_customer.unicode_value or '',
                           ticket.issue_type and ticket.issue_type.unicode_value or ''])


class QualityRaw(Quality):
    def __call__(self, parser, namespace, values, option_string):
        super(QualityRaw, self).__call__(parser, namespace, values, option_string)
        session = DBSession()
        with open(namespace.filename, 'wb') as ofile:
            writer = csv.writer(ofile, dialect='excel')
            writer.writerow(['CR ID', 'Customer', 'Duration in days'])
            for cr in session.query(CustomerRequest.id,
                                    CustomerRequest.project_id,
                                    Project.customer_id,
                                    Trac.trac_name).\
                              outerjoin(Project,
                                        CustomerRequest.project_id == Project.id).\
                              outerjoin(Trac,
                                        Project.id==Trac.project_id).distinct():

                tickets, Ticket, TicketCustom = tickets_for_cr(self.metadata, session, cr.trac_name, cr.id)
                tids = [t.id for t in tickets]
                if not tids:
                    continue

                entries = session.query(TimeEntry)\
                                 .filter_by(project_id=cr.project_id)\
                                 .filter(TimeEntry.ticket.in_(tids))\
                                 .filter(extract('year', TimeEntry.date) == namespace.year)

                if entries.count():
                    hours = timedelta_as_work_days(sum([a.hours for a in entries], timedelta()))
                    writer.writerow([cr.id, cr.customer_id, hours])


def main():
    """
    ./bin/quality_export etc/production.ini export_file.csv --report-name
    """

    if len(sys.argv) == 1:
        sys.argv.append('--help')

    parser = argparse.ArgumentParser(description='Quality export')
    DEFAULT_YEAR = date.today().year -1
    parser.set_defaults(year=DEFAULT_YEAR)

    parser.add_argument('configuration', action='store', help='path to the wsgi configuration ini file.')
    parser.add_argument('year', help='Report year. (default %s)' % DEFAULT_YEAR, action='store', type=int)
    #parser.add_argument('--filename', help='path to the output CSV file.')
    parser.add_argument('--google', help='google folder id.')
    parser.add_argument('--project', nargs=0, action=QualityProject, help='generate project quality report.')
    parser.add_argument('--cr', nargs=0, action=QualityCR, help='generate customer request quality report.')
    parser.add_argument('--ticket', nargs=0, action=QualityTicket, help='generate ticket quality report.')
    parser.add_argument('--raw', nargs=0, action=QualityRaw, help='generate raw quality report.')
    namespace = parser.parse_args()

    if namespace.google:
        client = create_client()
        for report in  ['project', 'cr', 'ticket', 'raw']:
            configuration = getattr(namespace, report)
            if configuration:
                upload_file(client, configuration['filename'],
                            'Report: %s' % report,
                            namespace.google)
                os.unlink(configuration['filename'])



from copy import deepcopy
import json

from sqlalchemy import not_
from pyramid import httpexceptions as exc
from pyramid_skins import SkinObject

from fa.bootstrap import actions

from penelope.core.models import dashboard
from penelope.core.forms import ModelView
from penelope.core.forms.renderers import ProjectRelationRenderer
from penelope.core.reports.all_entries import AllEntriesReport
from penelope.core.reports.views import ReportContext
from penelope.core.forms.wizard import Wizard


def configurate(config):
    #custom view for customer

    config.formalchemy_model_view('admin',
        request_method='GET',
        permission='metadata',
        name='metadata',
        attr='show',
        renderer='penelope.core.forms:templates/customer.pt',
        model='penelope.core.models.dashboard.Customer',
        view=CustomerModelView)

    config.formalchemy_model_view('admin',
        request_method='POST',
        permission='delete',
        name='delete',
        attr='delete',
        renderer='fa.bootstrap:templates/admin/edit.pt',
        model='penelope.core.models.dashboard.Customer',
        view=CustomerModelView)

    config.formalchemy_model_view('admin',
        renderer='penelope.core.forms:templates/customer_listing.pt',
        attr='datatable',
        context='pyramid_formalchemy.resources.ModelListing',
        request_method='GET',
        permission='customer_listing',
        model='penelope.core.models.dashboard.Customer',
        view=CustomerModelView)

    #custom view for adding a project to a customer
    config.formalchemy_model_view('admin',
        request_method='GET',
        permission='new',
        name='add_project',
        attr='add_project',
        renderer='penelope.core.forms:templates/new.pt',
        model='penelope.core.models.dashboard.Customer',
        view=CustomerModelView)

    config.formalchemy_model_view('admin',
        request_method='POST',
        permission='new',
        name='add_project',
        attr='add_project',
        renderer='penelope.core.forms:templates/new.pt',
        model='penelope.core.models.dashboard.Customer',
        view=CustomerModelView)

    #custom view for adding a project to a customer
    config.formalchemy_model_view('admin',
        request_method='GET',
        permission='new',
        name='wizard',
        attr='render',
        renderer='penelope.core.forms:templates/wizard.pt',
        model='penelope.core.models.dashboard.Customer',
        view=Wizard)
    #custom view for adding a project to a customer
    config.formalchemy_model_view('admin',
        request_method='POST',
        permission='new',
        name='wizard',
        attr='render',
        renderer='penelope.core.forms:templates/wizard.pt',
        model='penelope.core.models.dashboard.Customer',
        view=Wizard)

    #custom view for customer projects
    config.formalchemy_model_view('admin',
        request_method='GET',
        permission='view',
        name='',
        attr='customer_projects',
        renderer='penelope.core.forms:templates/customer_projects.pt',
        model='penelope.core.models.dashboard.Customer',
        view=CustomerModelView)

    #custom view for customer projects
    config.formalchemy_model_view('admin',
        request_method='GET',
        permission='time_entries',
        name='customer_time_entries',
        attr='customer_time_entries',
        model='penelope.core.models.dashboard.Customer',
        view=CustomerModelView)


add_project = actions.UIButton(id='add_project',
    content='Add project',
    permission='new',
    _class='btn btn-success',
    attrs=dict(href="request.fa_url('Customer', request.model_id, 'add_project')"))

add_project_wizard = actions.UIButton(id='add_project_wizard',
    content='Add project wizard',
    permission='new',
    _class='btn btn-success',
    attrs=dict(href="request.fa_url('Customer', request.model_id, 'wizard')"))

customer_tabs = actions.Actions(actions.TabAction("show",
    content="Customer",
    attrs=dict(href="request.fa_url('Customer', request.model_id, '')")),
    actions.TabAction("metadata",
        permission='metadata',
        content="Metadata",
        attrs=dict(href="request.fa_url('Customer', request.model_id, 'metadata')")),
    actions.TabAction("customer_time_entries",
        content="Time entires",
        permission='time_entries',
        attrs=dict(
            href="request.fa_url('Customer', request.model_id, 'customer_time_entries', customer_id=request.model_id)"))
    , )


class CustomerModelView(ModelView):
    actions_categories = ('buttons', 'tabs')

    defaults_actions = deepcopy(actions.defaults_actions)
    defaults_actions.update(show_buttons=actions.Actions(actions.edit))
    defaults_actions['customer_projects_buttons'] = actions.Actions(add_project, add_project_wizard)
    defaults_actions.update(show_tabs=customer_tabs)
    defaults_actions.update(customer_projects_tabs=customer_tabs)
    defaults_actions.update(customer_time_entries_tabs=customer_tabs)

    def add_project_url(self, *args, **kwargs):
        return self.project_url

    def add_project(self, *args, **kwargs):
        self.request.model_class = dashboard.Project
        self.request.model_name = dashboard.Project.__name__
        self.request.form_action = ['Customer', self.request.model_id, 'add_project']
        return self.new()

    @actions.action('customer_projects')
    def customer_projects(self, *args, **kwargs):
        context = self.context.get_instance()

        query = self.session.query(dashboard.Project).filter(dashboard.Project.customer_id == context.id)
        other_projects = self.request.query_factory(self.request, query).filter(not_(dashboard.Project.active))

        active_project_query = query.filter(dashboard.Project.active)
        active_projects = self.request.query_factory(self.request, active_project_query)

        self.context.get_model = lambda: dashboard.Project
        self.request.model_name = 'Project'

        page = self.get_page(collection=list(self.request.filter_viewables(other_projects)))
        fs = self.get_grid()
        fs = fs.bind(instances=page, request=self.request)
        fs.configure(readonly=True)
        del fs._render_fields['applications']
        del fs._render_fields['time_entries']
        del fs._render_fields['favorite_users']
        del fs._render_fields['creation_date']
        del fs._render_fields['modification_date']
        del fs._render_fields['author']
        del fs._render_fields['activated']
        del fs._render_fields['customer']
        del fs._render_fields['groups']
        del fs._render_fields['customer_requests']
        del fs._render_fields['completion_date']
        del fs._render_fields['test_date']
        del fs._render_fields['inception_date']
        del fs._render_fields['assistance_date']

        page = self.get_page(collection=list(self.request.filter_viewables(active_projects)))
        fs_active = self.get_grid()
        fs_active = fs_active.bind(instances=page, request=self.request)
        fs_active.configure(readonly=True)
        del fs_active._render_fields['applications']
        del fs_active._render_fields['time_entries']
        del fs_active._render_fields['favorite_users']
        del fs_active._render_fields['creation_date']
        del fs_active._render_fields['modification_date']
        del fs_active._render_fields['activated']
        del fs_active._render_fields['customer']
        del fs_active._render_fields['author']
        del fs_active._render_fields['groups']
        del fs_active._render_fields['customer_requests']
        del fs_active._render_fields['completion_date']
        del fs_active._render_fields['test_date']
        del fs_active._render_fields['assistance_date']
        del fs_active._render_fields['inception_date']

        return self.render(fs=fs, fs_active=fs_active)

    @actions.action('customer_time_entries')
    def customer_time_entries(self, *args, **kwargs):
        params = AllEntriesReport(self.context, self.request)()
        params.update(**self.render())
        params.update(request=self.request)
        report_context= ReportContext(self.request)
        params.update(context=report_context)
        return SkinObject('customer_time_entries')(**params)

    @actions.action('listing')
    def datatable(self, **kwargs):
        result = super(CustomerModelView, self).datatable(**kwargs)

        fs = result['fs']
        fs.configure(pk=True, readonly=True)
        viewed_customers = set(fs.rows.items)

        columns = ['name', 'projects', 'id']
        self.pick_columns(fs, columns)

        fs._render_fields['projects']._get_renderer = lambda: ProjectRelationRenderer(fs._render_fields['projects'])

        favorite_projects = set(self.request.authenticated_user.favorite_projects)
        favorite_customers = set(c.id
                                 for c in viewed_customers
                                 if favorite_projects.intersection(c.projects))
        active_customers = set(c.id
                               for c in viewed_customers
                               if any(p.activated for p in c.projects))

        return dict(result,
                    columns=columns,
                    js_active_customers=json.dumps(list(active_customers)),
                    js_favorite_customers=json.dumps(list(favorite_customers)))


    def delete(self):
        customer = self.context.get_instance()
        if customer.projects:
            request = self.request
            request.add_message(u'Customer has projects. Please remove them first.', type='danger')
            raise exc.HTTPFound(location=request.fa_url('Customer', customer.id))
        else:
            return self.force_delete()



import os
import sys
import transaction
import random
import beaker

from sqlalchemy import engine_from_config
from pyramid.paster import setup_logging
from pyramid.paster import bootstrap

from penelope.core.models.dashboard import Trac
from penelope.core.models.scripts import add_customer, add_customer_request, add_project, populate_time_entries, add_user, add_role
from penelope.core.models.dbsession import DBSession
from penelope.core.models import Base
from penelope.core.models.dashboard import GlobalConfig
from penelope.trac.populate import add_trac_tickets


beaker.cache.cache_regions.update(dict(calculate_matrix={'key_length':''}))


def usage(argv):
    cmd = os.path.basename(argv[0])
    print('usage: %s <config_uri>\n'
          '(example: "%s development.ini [--use-small]")' % (cmd, cmd)) 
    sys.exit(1)


def main(argv=sys.argv):
    if len(argv) > 3:
        usage(argv)
    if len(argv) == 3:
        use_small = True
    else:
        use_small = False
    config_uri = argv[1]
    setup_logging(config_uri)
    env = bootstrap('%s#dashboard'% config_uri)
    settings = env.get('registry').settings
    engine = engine_from_config(settings, 'sa.dashboard.')
    DBSession.configure(bind=engine)
    Base.metadata.bind = engine
    Base.metadata.create_all()


    with transaction.manager:
        random.seed(42)
        session = DBSession()
        session.add(GlobalConfig(id=1))

        users = [
                 add_user(session, u'john@example.com', fullname='John Smith'),
                 add_user(session, u'customer@example.com', fullname='Mr. Customer'),
        ]

        role_admin = add_role(session, 'administrator')
        role_admin.users.append(users[0])

        add_role(session, 'external_developer')
        add_role(session, 'internal_developer')
        add_role(session, 'secretary')
        role_customer = add_role(session, 'customer')
        role_customer.users.append(users[1])
        add_role(session, 'project_manager')

        if use_small:
            small_dummies(session,users,settings)
        else:
            full_dummies(session,users,settings)


def small_dummies(session, users, settings):
        customers = [
            add_customer(session, u'Mr. Customer'),
        ]
        session.flush()

        projects = []
        for project_name, customer_id in [
                (u'Ferrara', customers[0].id),
                (u'Bologna', customers[0].id),
            ]:
            projects.append(add_project(session, project_name, customer_id, None))
            session.flush()

        customer_requests = [
            add_customer_request(session, u"Want this",                     project_id='ferrara'),
            add_customer_request(session, u"Want that",                     project_id='bologna'),
        ]
        session.flush()

        all_tickets = {}
        for project in projects:
            all_tickets[project.id] = []
            # add_trac_to_project(project, config)
            trac_app = Trac(name='trac for project %s' % project.id)
                            # api_uri e' generata dinamicamente dall'evento before_insert di application
                            # api_uri = 'http://localhost:8081/trac/%s' % project.id,
            project.add_application(trac_app)
        session.flush()

        for cr in customer_requests:
            all_tickets[cr.project.id].extend(add_trac_tickets(cr.project, cr, settings, users))
        populate_time_entries(session, users, projects, all_tickets)


def full_dummies(session, users, settings):
        customers = [
            add_customer(session, u'Mr. Customer'),
            add_customer(session, u'Rick Astley'),
            add_customer(session, u'Lady Gaga'),
            add_customer(session, u'Tom Waits'),
        ]

        session.flush()

        projects = []
        for project_name, customer_id in [
                (u'Plone', customers[0].id),
                (u'Pyramid', customers[0].id),
                (u'Ferrara', customers[1].id),
                (u'Bologna', customers[1].id),
                (u'Verona', customers[2].id),
                (u'Genova', customers[2].id),
                (u'Modena', customers[3].id),
                (u'Batman Returns', customers[3].id),
                (u'Totoro 1', customers[1].id),
                (u'Totoro 2', customers[2].id),
                (u'Totoro 3', customers[3].id),
            ]:
            projects.append(add_project(session, project_name, customer_id, None))
            session.flush()


        customer_requests = [
            add_customer_request(session, u"Want this",                     project_id='plone'),
            add_customer_request(session, u"Want that",                     project_id='pyramid'),
            add_customer_request(session, u"Want something different",      project_id='ferrara'),
            add_customer_request(session, u"Doesn't really know",           project_id='ferrara'),
            add_customer_request(session, u"Will know when it's done",      project_id='bologna'),
            add_customer_request(session, u"A pony",                        project_id='verona'),
            add_customer_request(session, u"OMG double rainbow!!",          project_id='verona'),
            add_customer_request(session, u"A white kitten",                project_id='verona'),
            add_customer_request(session, u"A scary jabberwocky",           project_id='genova'),
            add_customer_request(session, u"A black kitten",                project_id='genova'),
            add_customer_request(session, u"A Facebook clone",              project_id='genova'),
            add_customer_request(session, u"Many unrelated things",         project_id='modena'),
            add_customer_request(session, u"The blink tag",                 project_id='modena'),
            add_customer_request(session, u"And one more thing..",          project_id='modena'),
            add_customer_request(session, u"Emacs emulation for EDLIN.COM", project_id='batman-returns'),
            add_customer_request(session, u"Totoro 1",                      project_id='totoro-1'),
            add_customer_request(session, u"Stuff for Totoro 2",            project_id='totoro-2'),
            add_customer_request(session, u"Totoro 3 again",                project_id='totoro-3'),
        ]

        session.flush()

        all_tickets = {}
        for project in projects:
            all_tickets[project.id] = []
            # add_trac_to_project(project, config)
            trac_app = Trac(name='trac for project %s' % project.id)
                            # api_uri e' generata dinamicamente dall'evento before_insert di application
                            # api_uri = 'http://localhost:8081/trac/%s' % project.id,
            project.add_application(trac_app)

        session.flush()

        for cr in customer_requests:
            all_tickets[cr.project.id].extend(add_trac_tickets(cr.project, cr, settings, users))
        populate_time_entries(session, users, projects, all_tickets)




# -*- coding: utf-8 -*-

import sqlalchemy as sa

from penelope.core.models import DBSession, Project, TimeEntry, CustomerRequest


class NullCustomerRequest(object):
    def __init__(self, project=None):
        self.project = project

    name = '(no request)'



def qry_active_projects():
    return DBSession.query(Project)\
            .filter(Project.activated==True)\
            .filter(Project.customer_id!=None)\
            .order_by(Project.name)


def filter_users_with_timeentries(users):
    with_te = set(row.author_id for row in DBSession.query(TimeEntry.author_id).group_by(TimeEntry.author_id))
    return [user for user in users if user.id in with_te]


def te_filter_by_customer_requests(customer_requests, request):
    """
    Returns a SQL Expression to filter time entries, that belong to the provided customer_requests.
    The returned expression can be applied to a query on the TimeEntry table.

    If customer_requests is empty, no filter is applied.
    """

    if not customer_requests:
        return sa.text('1=1')

    return TimeEntry.customer_request_id.in_(customer_requests)


def te_filter_by_contracts(contracts):
    if not contracts:
        return sa.text('1=1')

    # get customer requests for contracts
    customer_requests = DBSession().query(CustomerRequest.id).filter(CustomerRequest.contract_id.in_(contracts))
    return te_filter_by_customer_requests(customer_requests, None)




from fa.jquery.fanstatic_resources import fa_pyramid_js
from fa.jquery.renderers import RichTextFieldRenderer
from formalchemy import Field, fatypes
from formalchemy.exceptions import ValidationError
from formalchemy.fields import _query_options
from formalchemy.fields import HiddenFieldRenderer, SelectFieldRenderer
from plone.i18n.normalizer import idnormalizer
from pyramid.i18n import TranslationStringFactory
from pyramid.security import has_permission
from pyramid_formalchemy import events
from sqlalchemy.util import OrderedDict
from sqlalchemy import or_
from zope.interface import Interface

from penelope.core import PROJECT_ID_BLACKLIST
from penelope.core.forms.renderers import UrlRenderer
from penelope.core.forms.renderers import TicketRenderer
from penelope.core.lib.fa_fields import BigTextAreaFieldRenderer
from penelope.core.models import DBSession
from penelope.core.models.dashboard import TRAC, SVN, Application, Customer, \
        CustomerRequest, Group, Project, User, Contract, KanbanBoard, \
        GOOGLE_DOCS, TRAC_REPORT, GENERIC_APP, BACKLOG_PRIORITY_ORDER, \
        BACKLOG_MODIFICATION_ORDER
from penelope.core.models.tp import TimeEntry
from penelope.core.models.dublincore import DublinCore
from penelope.core.models.interfaces import IRoleable, ITimeEntry
from penelope.core.models.workflow import Workflow

_ = TranslationStringFactory('penelope')


@events.subscriber([Interface, events.IBeforeEditRenderEvent])
def before_generic_edit_render(context, event):
    fa_pyramid_js.need()


@events.subscriber([DublinCore, events.IAfterSyncEvent])
def after_dublincore_sync(context, event):
    context.request = event.request #we need request for user calculation


#Dublincore rendering events
@events.subscriber([DublinCore, events.IBeforeListingRenderEvent])
def before_dublincore_listing_render(context, event):
    fs = event.kwargs['fs']
    if not fs._render_fields.keys():
        fs.configure(readonly=fs.readonly)
    del fs._render_fields['creation_date']
    del fs._render_fields['modification_date']
    del fs._render_fields['author']


@events.subscriber([DublinCore, events.IBeforeEditRenderEvent])
def before_dublincore_edit_render(context, event):
    fs = event.kwargs['fs']
    if not fs._render_fields.keys():
        fs.configure()
    del fs._render_fields['creation_date']
    del fs._render_fields['modification_date']
    if has_permission('manage', context, event.request) and ITimeEntry.providedBy(context):
        pass
    else:
        del fs._render_fields['author']


#Role mapping events
@events.subscriber([IRoleable, events.IBeforeEditRenderEvent])
def before_role_render(context, event):
    fs = event.kwargs['fs']
    if not fs._render_fields.keys():
        fs.configure(readonly=fs.readonly)
    if not has_permission('manage_roles', context, event.request):
        del fs._render_fields['roles']


#Workflow events
@events.subscriber([Workflow, events.IBeforeRenderEvent])
def before_workflow_render(context, event):
    fs = event.kwargs['fs']
    if not fs._render_fields.keys():
        fs.configure(readonly=fs.readonly)
    fs._render_fields.pop('invoice_number', None)
    wf = fs._render_fields.pop('workflow_state')
    fs.append(wf)


@events.subscriber([Workflow, events.IBeforeEditRenderEvent])
def before_workflow_edit_render(context, event):
    fs = event.kwargs['fs']
    if not fs._render_fields.keys():
        fs.configure()
    del fs._render_fields['workflow_state']



#Project rendering events
@events.subscriber([Project, events.IBeforeRenderEvent])
def before_project_render(context, event):
    fs = event.kwargs['fs']
    if not fs._render_fields.keys():
        fs.configure(readonly=fs.readonly)
    del fs._render_fields['applications']
    del fs._render_fields['time_entries']
    del fs._render_fields['favorite_users']

@events.subscriber([Project, events.IBeforeEditRenderEvent])
def before_project_edit_render(context, event):
    fs = event.kwargs['fs']
    if not fs._render_fields.keys():
        fs.configure()
    fs.test_date.set(instructions=_(u'the date (if any) in which the Customer has officially accepted the project as "completed"'))
    fs.assistance_date.set(instructions=_(u'the date (if any) when the post-release assistance is ending'))
    fs.completion_date.set(instructions=_(u'the date in which the full invoicing of the developement phase is completed, excluding post-release assistance. Usually, it is the date of the final invoice.'))
    fs.append(fs.name.required())
    del fs._render_fields['customer_requests']
    del fs._render_fields['groups']
    pk_to_add = [a for a in fs._raw_fields() if a.name == 'id']
    fs.append(pk_to_add[0].set(renderer=HiddenFieldRenderer))

@events.subscriber([Project, events.IBeforeNewRenderEvent])
def before_project_new_render(context, event):
    """called after edit renderer"""
    bind_customer(context, event)
    fs = event.kwargs['fs']
    del fs._render_fields['customer']
    pk_to_add = [a for a in fs._raw_fields() if a.name == 'id']
    items = list(tuple(fs._render_fields.iteritems()))
    items.insert(0, (pk_to_add[0].name, pk_to_add[0]))
    fs._render_fields = OrderedDict(items)
    fs._render_fields['id'].validators = []
    fs._render_fields['id']._renderer = None


@events.subscriber([Project, events.IBeforeValidateEvent])
def before_project_validated(context, event):
    """called before validation"""
    def my_validator(fs):
        if not fs.id.value:
            project_id = idnormalizer.normalize(fs.name.value)
        else:
            project_id = fs.id.value

        if project_id.lower() in PROJECT_ID_BLACKLIST:
            msg = _('${fs_name_value} is a restricted name! Please choose '
                    'another project name or provide unique ID!',
                    mapping={'fs_name_value': project_id})
            raise ValidationError(msg)

        project = DBSession().query(Project).get(project_id)
        if project and project != fs.model:
            msg = _('${fs_name_value} already exists! Please choose another '
                    'project name or provide unique ID!',
                    mapping={'fs_name_value': project_id})
            raise ValidationError(msg)

        if fs.activated.value == False and project:
            active_contracts = [a for a in project.contracts if a.active]
            if active_contracts:
                msg = _('You cannot deactivate project! It has uncompleted '
                        'contracts!')
                raise ValidationError(msg)

    event.fs.validator = my_validator


#Customer rendering events
@events.subscriber([Customer, events.IBeforeEditRenderEvent])
def before_customer_render(context, event):
    fs = event.kwargs['fs']
    #fs.projects.set(renderer=autocomplete_relation(filter_by='name'))
    fs.append(fs.name.required())
    del fs._render_fields['projects']


#User listing
@events.subscriber([User, events.IBeforeListingRenderEvent])
def before_user_listing_render(context, event):
    fs = event.kwargs['fs']
    if not fs._render_fields.keys():
        fs.configure(readonly=fs.readonly)
    del fs._render_fields['mobile']
    del fs._render_fields['phone']
    del fs._render_fields['groups']
    del fs._render_fields['favorite_projects']


#User rendering events
@events.subscriber([User, events.IBeforeRenderEvent])
def before_user_render(context, event):
    fs = event.kwargs['fs']
    if not fs._render_fields.keys():
        fs.configure(readonly=fs.readonly)
    del fs._render_fields['gdata_auth_token']
    del fs._render_fields['openids']
    del fs._render_fields['salt']
    del fs._render_fields['password']
    del fs._render_fields['active']
    del fs._render_fields['kanban_boards']


@events.subscriber([User, events.IBeforeEditRenderEvent])
def before_user_edit_render(context, event):
    fs = event.kwargs['fs']
    if not fs._render_fields.keys():
        fs.configure(readonly=fs.readonly)

    del fs._render_fields['favorite_projects']
    del fs._render_fields['groups']

    if not has_permission('manage', context, event.request):
        del fs._render_fields['project_manager']

    fs.append(fs.fullname.required())
    fs.append(fs.email.required())


#Group rendering events
@events.subscriber([Group, events.IBeforeEditRenderEvent])
def before_group_render(context, event):
    bind_project(context, event)
    fs = event.kwargs['fs']
    if not fs._render_fields.keys():
        fs.configure(readonly=fs.readonly)
    del fs._render_fields['project']


@events.subscriber([Application, events.IBeforeRenderEvent])
def before_application_render(context, event):
    fs = event.kwargs['fs']
    if not fs._render_fields.keys():
        fs.configure(readonly=fs.readonly)
    del fs._render_fields['svn_name']
    del fs._render_fields['trac_name']
    del fs._render_fields['acl']
    fs.api_uri.set(renderer=UrlRenderer) 


@events.subscriber([Application, events.IBeforeValidateEvent])
def before_application_validated(context, event):
    """called before validation"""
    def my_validator(fs):
        app_type = fs.application_type.value
        if app_type != TRAC and app_type != SVN:
            if not fs.api_uri.value:
                msg = _('You have choosen ${app_type} as your application type. Please provide api uri.',
                        mapping={'app_type': app_type})
                raise ValidationError(msg)
    event.fs.validator = my_validator


@events.subscriber([Application, events.IBeforeEditRenderEvent])
def before_application_edit_render(context, event):
    fs = event.kwargs['fs']
    if not fs._render_fields.keys():
        fs.configure(readonly=fs.readonly)
    fs.description.set(renderer=RichTextFieldRenderer(use='tinymce', theme='simple'))
    fs.append(fs.name.required())
    del fs._render_fields['project']
    fs.append(Field('application_type', type=fatypes.String))
    fs.application_type.set(renderer=SelectFieldRenderer,
            options=[('Trac', TRAC),
                     ('SVN', SVN),
                     ('Trac Report', TRAC_REPORT),
                     ('Google Docs', GOOGLE_DOCS),
                     ('Generic', GENERIC_APP)])
    bind_project(context, event)
    [fs.append(fs._render_fields.pop(a)) for a in fs._render_fields if a != 'name']


@events.subscriber([Application, events.IBeforeNewRenderEvent])
def before_application_new_render(context, event):
    fs = event.kwargs['fs']
    if not fs._render_fields.keys():
        fs.configure(readonly=fs.readonly)
    del fs._render_fields['position']
    fs.api_uri.metadata['instructions'] = _(u'Please provide application uri. If you choose trac or svn - leave this field empty.')


@events.subscriber([Contract, events.IBeforeRenderEvent])
def before_contract_render(context, event):
    bind_project(context, event)
    fs = event.kwargs['fs']
    if not fs._render_fields.keys():
        fs.configure(readonly=fs.readonly)
    del fs._render_fields['project']
    del fs._render_fields['customer_requests']


#Contract listing
@events.subscriber([Contract, events.IBeforeListingRenderEvent])
def before_contract_listing_render(context, event):
    fs = event.kwargs['fs']
    if not fs._render_fields.keys():
        fs.configure(readonly=fs.readonly)
    del fs._render_fields['description']
    del fs._render_fields['project_id']


@events.subscriber([Contract, events.IBeforeEditRenderEvent])
def before_contract_editrender(context, event):
    bind_project(context, event)
    fs = event.kwargs['fs']
    fs.contract_number.set(instructions=_(u'Something like number/year'))
    fs.amount.set(instructions=_(u'Contract amount in EUR'))
    if not fs._render_fields.keys():
        fs.configure(readonly=fs.readonly)
    fs.description.set(renderer=RichTextFieldRenderer(use='tinymce', theme='simple'))
    fs.append(fs.name.required())
    for field in ['name', 'contract_number', 'amount', 'days', 'start_date', 'end_date', 'description']:
        fs.append(fs._render_fields.pop(field))


@events.subscriber([Contract, events.IBeforeValidateEvent])
def before_contract_validated(context, event):
    """called before validation"""
    def my_validator(fs):
        contract_id = idnormalizer.normalize('%s_%s' % (fs.project_id.value, fs.name.value))

        if contract_id.lower() in PROJECT_ID_BLACKLIST:
            msg = _('${fs_name_value} is a restricted name! Please choose '
                    'another contract name!',
                    mapping={'fs_name_value': fs.name.value})
            raise ValidationError(msg)

        contract = DBSession().query(Contract).get(contract_id)
        if contract and contract != fs.model:
            msg = _('${fs_name_value} already exists! Please choose another '
                    'contract name!',
                    mapping={'fs_name_value': fs.name.value})
            raise ValidationError(msg)
    event.fs.validator = my_validator


#Customer request rendering events
@events.subscriber([CustomerRequest, events.IBeforeEditRenderEvent])
def before_customerrequest_editrender(context, event):
    bind_project(context, event)
    fs = event.kwargs['fs']
    if not fs._render_fields.keys():
        fs.configure(readonly=fs.readonly)
    fs.description.set(renderer=RichTextFieldRenderer(use='tinymce', theme='simple'))
    fs.append(fs.name.required())
    del fs._render_fields['project']

    if context.active:
        q = DBSession().query(fs.contract.relation_type())\
                       .filter_by(project_id=context.project_id)\
                       .filter(or_(fs.contract.relation_type().workflow_state=='draft',
                                   fs.contract.relation_type().workflow_state=='active'))\
                       .order_by('name')
    else:
        q = DBSession().query(fs.contract.relation_type())\
                       .filter_by(project_id=context.project_id)\
                       .order_by('name')

    fs.contract.render_opts['options'] = _query_options(q)

    [fs.append(fs._render_fields.pop(a)) for a in fs._render_fields if a != 'name']


@events.subscriber([CustomerRequest, events.IBeforeRenderEvent])
def before_customerrequest_render(context, event):
    bind_project(context, event)
    fs = event.kwargs['fs']
    if not fs._render_fields.keys():
        fs.configure(readonly=fs.readonly)
    del fs._render_fields['estimations']
    del fs._render_fields['time_entries']
    del fs._render_fields['uid']
    del fs._render_fields['project_id']
    safe_project = DBSession().query(fs.project.relation_type()).get(context.project_id)
    if not event.request.has_permission('add_contract', safe_project):
        del fs._render_fields['contract']
    if fs.readonly:
        fs.append(Field('estimation_days', type=fatypes.Float))
        fs.estimation_days._value = context.estimation_days

@events.subscriber([KanbanBoard, events.IBeforeRenderEvent])
def before_kanban_render(context, event):
    fs = event.kwargs['fs']
    if not fs._render_fields.keys():
        fs.configure(readonly=fs.readonly)
    del fs._render_fields['json']
    del fs._render_fields['acl']


@events.subscriber([KanbanBoard, events.IBeforeEditRenderEvent])
def before_kanban_edit_render(context, event):
    fs = event.kwargs['fs']
    if not fs._render_fields.keys():
        fs.configure(readonly=fs.readonly)
    q = event.request.filter_viewables(DBSession().\
                      query(fs.projects.relation_type()).filter(fs.projects.relation_type().active).order_by('name'))
    fs.projects.render_opts['options'] = _query_options(q)
    fs.backlog_query.set(renderer=BigTextAreaFieldRenderer)
    fs.backlog_query.set(instructions=_(u'This is the SQL WHERE statement.<br/>'
                                      u'The default query equals to: <code>ticket.owner=\'%s\' AND ticket.status!=\'closed\'</code>.<br/>'
                                      u'Available fields: <code>ticket.owner, ticket.status, ticket.id, ticket.priority, ticket.reporter, customerrequest.value, probabilita.value</code>' % event.request.authenticated_user.email))
    if not fs.backlog_query.value:
        context.backlog_query = "owner='%s' AND status!='closed'" % event.request.authenticated_user.email
    fs.backlog_order.set(renderer=SelectFieldRenderer,
                options=[('Priority', BACKLOG_PRIORITY_ORDER),
                         ('Modification', BACKLOG_MODIFICATION_ORDER)])


#TimeEntry events
@events.subscriber([TimeEntry, events.IBeforeShowRenderEvent])
def before_timeentry_show_render(context, event):
    fs = event.kwargs['fs']
    fs.ticket.set(renderer=TicketRenderer(fs.ticket))


@events.subscriber([TimeEntry, events.IBeforeEditRenderEvent])
def before_timeentry_edit_render(context, event):
    fs = event.kwargs['fs']
    if not fs._render_fields.keys():
        fs.configure(readonly=fs.readonly)
    fs.description.set(renderer=BigTextAreaFieldRenderer)
    del fs._render_fields['start']
    del fs._render_fields['end']
    del fs._render_fields['tickettype']

    q = DBSession().query(fs.customer_request.relation_type()).filter_by(project_id=context.project_id).order_by('name')
    fs.customer_request.render_opts['options'] = _query_options(q)
    #remove location required validator
    if not fs.location.value:
        fs.location.model.location = u'RedTurtle'

    fs.append(fs.date.required())
    fs.append(fs.ticket.required())


@events.subscriber([TimeEntry, events.IBeforeListingRenderEvent])
def before_timeentry_listing_render(context, event):
    fs = event.kwargs['fs']
    if not fs._render_fields.keys():
        fs.configure(readonly=fs.readonly)
    del fs._render_fields['start']
    del fs._render_fields['end']
    #stupid sorter
    [fs.append(fs._render_fields.pop(a)) for a in fs._render_fields if a != 'description']


def bind_project(context, event):
    fs = event.kwargs['fs']
    if event.request.model_instance.__class__ is Project:
        fs.model.project_id = event.request.model_id
    fs.project_id.is_raw_foreign_key = False
    fs.project_id.set(renderer=HiddenFieldRenderer)
    fs.append(fs.project_id)


def bind_customer(context, event):
    fs = event.kwargs['fs']
    if event.request.model_instance.__class__ is Customer:
        fs.model.customer_id = event.request.model_id
    fs.customer_id.is_raw_foreign_key = False
    fs.customer_id.set(renderer=HiddenFieldRenderer)
    fs.append(fs.customer_id)


class AfterEntryCreatedEvent(object):
    """A search entry was created"""

    def __init__(self, entry, timeentry):
        self.entry = entry
        self.timeentry = timeentry




import logging
import re

import zope.component
from zope.interface import Interface

from webhelpers.html import HTML, literal
from pyramid_formalchemy.actions import Actions
from pyramid.renderers import get_renderer
from fa.bootstrap.actions import TabsActions, TabAction, UIButton

from penelope.core.interfaces import ISidebar
from penelope.core.lib.helpers import unicodelower
from penelope.core.models.interfaces import IProjectRelated, ICustomerRequest, IApplication#, IKanbanBoard

gsm = zope.component.getGlobalSiteManager()

log = logging.getLogger(__name__)


class Button(UIButton):
    body = '''<a class="${_class}" tal:attributes="%(attributes)s">${content}</a>'''



def safe_fa_url(*args):
    params = [ "'%s'" % arg for arg in args ]
    for arg in args:
        if isinstance(arg, int):
            continue
        if not re.match('^[\'a-z0-9_\-]+$', arg.lower()):
            log.error('Prevented unsafe arguments to eval: %s' % params)
            return "''"
    ret = 'request.fa_url(%s)' % ', '.join(params)
    return ret




class SidebarAction(TabAction):
    body = u'''<li class="${action.li_class(request)}">
                   <a tal:condition="not no_link"
                      tal:attributes="%(attributes)s">${content}</a>
                    <tal:nolink tal:condition="no_link">${content}</tal:nolink>
               </li>'''

    def isActive(self, request):
        if not self.attrs.get('href'):
            return False
        for _id in self.rcontext.get('children', ()):
            if _id in request.matchdict.get('traverse'):
                return True
        return request.path_url.strip('/') == eval(self.attrs['href']).strip('/')

    def li_class(self, request):
        return self.isActive(request) and 'active' or ''

    def render(self, request):
        if not self.rcontext.get('no_link'):
            self.rcontext['no_link'] = False
        return super(SidebarAction, self).render(request)


class DividerSidebarAction(SidebarAction):
    body = u'''<li class="divider"></li>'''


class HeaderSidebarAction(Actions, SidebarAction):
    body = u'''<li class="nav-header ${action.li_class(request)}">
                    <a tal:condition="not no_link" tal:attributes="%(attributes)s">${content}</a>
                    <tal:nolink tal:condition="no_link">${content}</tal:nolink>
                    <div class="nav-header-icons">${items}</div>
               </li>'''

    def __init__(self, *args, **kwargs):
        Actions.__init__(self)
        SidebarAction.__init__(self, *args, **kwargs)

    def render(self, request, **kwargs):
        items = Actions.render(self, request, **kwargs)
        self.rcontext.update(items=literal(items))
        return SidebarAction.render(self, request)


class SidebarActions(TabsActions):
    body='''<ul class="nav nav-list">${items}</ul>'''


class EmptyActions(TabsActions):
    body='''${items}'''


class SidebarRenderer(object):
    def __init__(self, context):
        self.context = context
        self.actions = SidebarActions()

    def render(self, request):
        return None

gsm.registerAdapter(SidebarRenderer, (Interface,), ISidebar)


class ProjectSidebarRenderer(SidebarRenderer):

    def __init__(self, *args, **kwargs):
        super(ProjectSidebarRenderer, self).__init__(*args, **kwargs)
        self.applications = EmptyActions()
        self.crs = EmptyActions()
        self.contracts = EmptyActions()

    def add_base_actions(self, request):
        project = self.context.project

        #Project name
        prj = HeaderSidebarAction('project_home',
                    content=project.name,
                    permission='view',
                    no_link=True)

        prj.append(Button(id='view',
                          content=literal('<i class="icon-info-sign icon-white"></i>'),
                          _class='btn btn-info btn-mini',
                          permission='edit',
                          attrs=dict(href=safe_fa_url('Project', project.id, 'metadata'),
                                     title="'Project metadata'")))

        if not request.authenticated_user in project.favorite_users:
            prj.append(Button(id='add_as_favorite',
                              content=literal('<i class="icon-star icon-white"></i>'),
                              permission='view',
                              _class='btn btn-success btn-mini',
                              attrs=dict(href=safe_fa_url('Project', project.id, 'toggle_favorite'),
                                         title="'Add as favorite'")))
        else:
            prj.append(Button(id='remove_as_favorite',
                              content=literal('<i class="icon-star icon-white"></i>'),
                              _class='btn btn-danger btn-mini',
                              permission='view',
                              attrs=dict(href=safe_fa_url('Project', project.id, 'toggle_favorite'),
                                         title="'Remove from favorites'")))
        prj.append(Button(id='project_edit',
                          content=literal('<i class="icon-cog icon-white"></i>'),
                          permission='edit',
                          _class='btn btn-warning btn-mini',
                          attrs=dict(href=safe_fa_url('Project', project.id, 'edit'),
                                     title="'Project configuration'")))
        self.actions.append(prj)
        self.actions.append(DividerSidebarAction('div'))

        #Documentation
        docs = HeaderSidebarAction('documentation',
                        content=u'Documentation',
                        permission='view',
                        no_link=True)
        docs.append(Button(id='add',
                           content=literal('<i class="icon-plus-sign icon-white"></i>'),
                           _class='btn btn-success btn-mini',
                           permission='new',
                           attrs=dict(href=safe_fa_url('Project', project.id, 'add_application'),
                                      title="'Add application'")))
        self.actions.append(docs)
        self.actions.append(self.applications)
        self.actions.append(SidebarAction('list_all_docs',
                                          content=u'List all...',
                                          permission='view',
                                          attrs=dict(href=safe_fa_url('Project', project.id, 'applications'))))
        self.actions.append(DividerSidebarAction('div'))

        #Customer request
        cr = HeaderSidebarAction('customer_requests',
                      content=u'Customer requests',
                      permission='view',
                      no_link=True)
        cr.append(Button(id='add',
                      content=literal('<i class="icon-plus-sign icon-white"></i>'),
                      _class='btn btn-success btn-mini',
                      permission='edit',
                      attrs=dict(href=safe_fa_url('Project', project.id, 'add_customer_request'),
                                 title="'Add customer request'")))
        self.actions.append(cr)
        self.actions.append(self.crs)
        self.actions.append(SidebarAction('list_all_cr',
                                          content=u'List all...',
                                          permission='view',
                                          attrs=dict(href=safe_fa_url('Project', project.id, 'customer_requests'))))

        #Contract
        if request.has_permission('view_contracts', self.context.project):
            contract = HeaderSidebarAction('contracts',
                        content=u'Contracts',
                        permission='view',
                        no_link=True)
            if request.has_permission('add_contracts', self.context.project):
                contract.append(Button(id='add',
                            content=literal('<i class="icon-plus-sign icon-white"></i>'),
                            _class='btn btn-success btn-mini',
                            permission='view',
                            attrs=dict(href=safe_fa_url('Project', project.id, 'add_contract'),
                                        title="'Add contract'")))
            self.actions.append(contract)
            self.actions.append(self.contracts)
            self.actions.append(SidebarAction('list_all_contracts',
                                            content=u'List all...',
                                            permission='view',
                                            attrs=dict(href=safe_fa_url('Project', project.id, 'contracts'))))
            self.actions.append(DividerSidebarAction('div'))

        #Permissions
        if request.has_permission('view_groups', self.context.project):
            permissions = HeaderSidebarAction('permissions',
                                content=u'Permissions',
                                permission='view',
                                no_link=True)

            permissions.append(Button(id='view',
                                content=literal('<i class="icon-info-sign icon-white"></i>'),
                                _class='btn btn-info btn-mini',
                                permission='view',
                                attrs=dict(href=safe_fa_url('Project', project.id, 'configuration'),
                                            title="'View groups'")))

            if request.has_permission('add_groups', self.context.project):
                permissions.append(Button(id='add',
                                        content=literal('<i class="icon-plus-sign icon-white"></i>'),
                                        _class='btn btn-success btn-mini',
                                        permission='view',
                                        attrs=dict(href=safe_fa_url('Project', project.id, 'add_group'),
                                                    title="'Add group'")))
            self.actions.append(permissions)

    def add_document_actions(self, only_trac=True):
        project = self.context.project
        if only_trac:
            apps = project.tracs
        else:
            apps = project.applications

        request = self.context.request

        if not request:
            return

        for app in request.filter_viewables(apps):
            if app.application_type == 'trac':
                href = "'%s'" % app.api_uri
            else:
                href = safe_fa_url('Application', app.id)

            self.applications.append(SidebarAction('app_%s' % app.id,
                                     content=literal(u'<i class="%s"></i> %s' % (app.get_icon(), app.name)),
                                     permission='view',
                                     attrs=dict(href=href)))

    def add_cr_actions(self, only_active=True):
        project = self.context.project
        for cr in sorted(project.customer_requests, key=unicodelower):
            if only_active:
                if cr.workflow_state != 'estimated':
                    continue

            icon = 'icon-time'
            self.crs.append(SidebarAction('cr_%s' % cr.id,
                                          content=literal(u'<i class="%s"></i> %s' % (icon, cr)),
                                          permission='view',
                                          attrs=dict(href=safe_fa_url('CustomerRequest', cr.id))))

    def add_contract_actions(self):
        project = self.context.project
        for cr in sorted(project.contracts, key=unicodelower):
            icon = 'icon-file'
            self.contracts.append(SidebarAction('cr_%s' % cr.id,
                                          content=literal('<i class="%s"></i> %s' % (icon, cr.name)),
                                          permission='view',
                                          attrs=dict(href=safe_fa_url('Contract', cr.id))))

    def render(self, request):
        project = self.context.project
        action = getattr(request, 'action', '')
        self.add_base_actions(request)
        self.add_document_actions(only_trac = action != 'applications')
        self.add_cr_actions(only_active = action != 'customer_requests')
        self.add_contract_actions()
        actions = self.actions.render(request)

        template =  get_renderer('penelope.core.forms:templates/project_sidebar.pt').implementation()
        return template(project=project,
                        actions=actions,
                        request=request)

gsm.registerAdapter(ProjectSidebarRenderer, (IProjectRelated,), ISidebar)


class CRSidebarRenderer(ProjectSidebarRenderer):
    def render(self, request):
        project = self.context.project

        self.add_base_actions(request)
        self.add_document_actions()
        self.add_cr_actions(only_active=False)
        actions = self.actions.render(request)

        template =  get_renderer('penelope.core.forms:templates/project_sidebar.pt').implementation()
        return template(project=project,
                        actions=actions,
                        request=request)

gsm.registerAdapter(CRSidebarRenderer, (ICustomerRequest,), ISidebar)


class AppSidebarRenderer(ProjectSidebarRenderer):
    def render(self, request):
        project = self.context.project

        self.add_base_actions(request)
        self.add_document_actions(only_trac=False)
        self.add_cr_actions()
        actions = self.actions.render(request)

        template =  get_renderer('penelope.core.forms:templates/project_sidebar.pt').implementation()
        return template(project=project,
                        actions=actions,
                        request=request)

gsm.registerAdapter(AppSidebarRenderer, (IApplication,), ISidebar)


class ManageSidebarRenderer(SidebarRenderer):

    def render(self, request):
        #Customer request
        board = HeaderSidebarAction('kanban_boards',
                      content=u'Boards',
                      permission='view',
                      no_link=True)
        board.append(Button(id='add',
                      content=literal('<i class="icon-plus-sign icon-white"></i>'),
                      _class='btn btn-success btn-mini',
                      permission='new',
                      attrs=dict(href=safe_fa_url('KanbanBoard', 'new'),
                                 title="'Add new board'")))
        self.actions.append(board)
        for kanban in request.authenticated_user.kanban_boards:
            icon = 'icon-time'
            self.actions.append(SidebarAction('board_%s' % kanban.id,
                                              content=literal(u'<i class="%s"></i> %s' % (icon, kanban)),
                                              permission='view',
                                              attrs=dict(href=safe_fa_url('KanbanBoard', kanban.id))))

        actions = self.actions.render(request)
        template =  get_renderer('penelope.core.forms:templates/project_sidebar.pt').implementation()
        return template(actions=actions, request=request)

#gsm.registerAdapter(ManageSidebarRenderer, (IKanbanBoard,), ISidebar)



"""
The DBSession is made available to model classes outside of penelope.core.models.__init__
to help avoid circular imports.
"""

from sqlalchemy.orm import scoped_session, sessionmaker
from zope.sqlalchemy import ZopeTransactionExtension

DBSession = scoped_session(sessionmaker(extension=ZopeTransactionExtension()))



# -*- coding: utf-8 -*-
from pyramid.httpexceptions import HTTPFound, HTTPOk
from pyramid.exceptions import Forbidden
from pyramid.renderers import get_renderer
from pyramid_skins import SkinObject
from dateutil.rrule import DAILY, rrule, SA, SU
from urlparse import urlparse, parse_qsl
from urllib import urlencode

from gdata.auth import AuthSubToken
from gdata.gauth import AuthSubToken as gAuthSubToken
from gdata.service import RequestError
from gdata.auth import generate_auth_sub_url
from gdata.docs.service import DocsService
from gdata.spreadsheet.service import SpreadsheetsService
from gdata.docs.client import DocsClient
from gdata.calendar.client import CalendarClient
from gdata.calendar.service import CalendarService

SCOPES = ['https://docs.google.com/feeds/', 'https://www.google.com/calendar/feeds/', 'https://spreadsheets.google.com/feeds/']


def fallback_tween_factory(handler, registry):
    def fallback_tween(request):
        try:
            response = handler(request)
        except GracefulFallback, e:
            qs_chr = request.query_string and '&' or '?'
            force_url = '%s%sforce=true' % (request.url, qs_chr)
            params = {'context':e.context,
                      'uri': e.kw.get('uri'),
                      'force_url': force_url,
                      'request':request}

            if hasattr(request.root,'fa_url'):
                #fa context
                params['main_template'] = get_renderer('penelope.core:skins/main_template.pt').implementation()
                params['main'] = get_renderer('penelope.core.forms:templates/master.pt').implementation()
                response = SkinObject('fallback_admin')(**params)
            else:
                response = SkinObject('fallback')(**params)
        return response
    return fallback_tween


def get_working_days(date_range, date_events_array):
    """
    It returns a dict of working day for each calendar.

    something like :
    {'%23contacts': 25, 'nicola.senno': 22, 'irene.capatti': 25, 'it.italian%23holiday': 25, 'amleczko': 24}
    """
    day_off = []
    start = date_range[0]
    end = date_range[1]

    for event in date_events_array:
        day_off.append(set(rrule(DAILY, dtstart=event[0], until=event[1])))

    all_dates = set(rrule(DAILY, dtstart=start, until=end))
    weekend = set(rrule(DAILY, dtstart=start, until=end, byweekday=(SA,SU)))
    all_no_weekend = all_dates.difference(weekend)
    merged = []

    for event_set in day_off:
        for event_date in event_set:
            merged.append(event_date)

    return  len(all_no_weekend.difference(set(merged)))


class AuthSubURLRedirect(HTTPFound):
    "Wraper for HTTPFound"
    def __init__(self, next_url, scopes, domain='default'):
        url = str(generate_auth_sub_url(next_url, scopes, domain=domain))
        super(AuthSubURLRedirect, self).__init__(url)


class GracefulFallback(HTTPOk):
    def __init__(self, context, **kw):
        self.context = context
        self.kw = kw
        super(GracefulFallback, self).__init__()


class GDataAuthDecorator(object):
    """ Decorator class for gdata auth."""

    def __init__(self, client, service):
        self.client =  client
        self.service = service
        self.scopes = SCOPES

    def _perform_auth(self, method, context, request, **kw):
        "Internal method"

        if not getattr(request, 'gclient', None):
            request.gclient = {}
        if not getattr(request, 'gservice', None):
            request.gservice = {}

        request.gclient[self.client.__class__.__name__] = self.client
        request.gservice[self.service.__class__.__name__] = self.service

        identity = request.environ.get('repoze.who.identity')
        if not identity:
            raise Forbidden

        user = identity.get('user')
        user_domains = user.email_domains
        if user_domains: 
            user_domain = user_domains[0]
        else:
            user_domain = 'default'
        next_url = request.url

        request_token = request.params.get('token')
        if request_token:
            authsub_token = AuthSubToken(scopes=self.scopes)
            authsub_token.set_token_string(request_token)
            self.service.auth_token = authsub_token
            self.service.UpgradeToSessionToken(token=authsub_token)
            user.gdata_auth_token = self.service.auth_token.get_token_string()

            #cleanup qs
            qs = dict(parse_qsl(urlparse(next_url).query))
            try:
                del qs['auth_sub_scopes']
                del qs['token']
                del qs['force']
            except KeyError: pass

            next_url = '%s?%s' % (request.path_url, urlencode(qs))
            raise HTTPFound(next_url)

        authsub_token = AuthSubToken(user.gdata_auth_token, scopes=self.scopes)
        self.service.SetAuthSubToken(user.gdata_auth_token)

        try:
            self.service.AuthSubTokenInfo()
        except RequestError:
            user.gdata_auth_token = None
            raise AuthSubURLRedirect(next_url, self.scopes, domain=user_domain)

        self.client.auth_token = gAuthSubToken(user.gdata_auth_token, scopes=self.scopes)
        return method(context, request, **kw)

    def auth_graceful(self, method):
        """
        Perform full gdata authsub authorization check.
        Fallback with URL if missing token

        Example:

            decorator = GDataAuthDecorator(GDClient(), GDService())

            @decorator.auth_graceful
            def get_file(context, request):
                return Response('OK')
        """
        def check_auth(context, request, **kw):
            """ perform authsub authorization """
            if request.params.get('force',False):
                return self._perform_auth(method, context, request, **kw)
            try:
                return self._perform_auth(method, context, request, **kw)
            except AuthSubURLRedirect:
                #gracefull fallback to method
                raise GracefulFallback(context, **kw)
        return check_auth

    def auth_required(self, method):
        """
        Perform full gdata authsub authorization check.
        FORCE authorization if missing.

        Example:

            decorator = GDataAuthDecorator(GDClient(), GDService())

            @decorator.auth_required
            def get_file(context, request):
                ...
                return Response('OK')
        """

        def check_auth(context, request, **kw):
            """ perform authsub authorization """
            return self._perform_auth(method, context, request, **kw)

        return check_auth

documents = GDataAuthDecorator(DocsClient(), DocsService())
spreadsheets = GDataAuthDecorator(DocsClient(), SpreadsheetsService())
calendar = GDataAuthDecorator(CalendarClient(), CalendarService())



import jsonrpc.proxy
import unittest2
import transaction
import datetime

from penelope.core.models.dashboard import User, Customer, CustomerRequest, Application, Project
from penelope.core.models import DBSession


class TestJSONRPCProxy(unittest2.TestCase):

    def setUp(self):
        self.proxy = jsonrpc.proxy.JSONRPCProxy('', path='/apis/json/dashboard')

    def test_get_user_by_attribute(self):
        session = DBSession()
        user = User(email=u'user@dummy.it')
        session.add(user)
        transaction.commit()
        #Get the user by email
        resp = self.proxy.get_user_by_email('user@dummy.it')
        self.assertEqual(resp['email'], u'user@dummy.it')

    def test_get_user_by_openid(self):
        session = DBSession()
        user = User(email=u'other_user@dummy.it')
        user.add_openid(u'other_user@dummy.it')
        session.add(user)
        transaction.commit()

        resp = self.proxy.get_user_by_openid('other_user@dummy.it')
        self.assertTrue(u'other_user@dummy.it' in resp['openids'])

    def test_get_customer_by_name(self):
        customer_name = u'A rich customer'
        session = DBSession()
        customer = Customer(name=customer_name)
        session.add(customer)
        transaction.commit()

        resp = self.proxy.get_customer_by_name(customer_name)
        self.assertEqual(resp['name'], customer_name)

    def test_get_project_by_name(self):
        project_name = u'A nice project'
        customer_name = u'A good customer'
        session = DBSession()
        project = Project(name=project_name)
        customer = Customer(name=customer_name)
        customer.add_project(project)
        session.add(customer)
        transaction.commit()

        resp = self.proxy.get_project_by_name(project_name)
        self.assertEqual(resp['name'], project_name)

    def test_project_attributes(self):
        project_name = u'Another nice project'
        customer_name = u'A very good customer'
        session = DBSession()
        project = Project(name=project_name)
        customer = Customer(name=customer_name)
        customer.add_project(project)
        applications1 = Application(name=u'Trac')
        applications2 = Application(name=u'Svn')
        customerR1 = CustomerRequest(name=u'A bad request')
        customerR2 = CustomerRequest(name=u'A good request')
        project.add_application(applications1)
        project.add_application(applications2)
        project.add_customer_request(customerR1)
        project.add_customer_request(customerR2)
        session.add(customer)
        transaction.commit()

        resp = self.proxy.get_project_by_name(project_name)
        self.assertTrue(u'Trac' in resp['applications'])
        self.assertTrue(u'another-nice-project_1' in [item for sublist in resp['customer_requests'] for item in sublist])
        self.assertTrue(u'another-nice-project_2' in [item for sublist in resp['customer_requests'] for item in sublist])
        self.assertTrue(2, len(resp['customer_requests']))
        self.assertTrue(2, len(resp['applications']))

    def test_not_found_user(self):
        resp = self.proxy.get_user_by_login('nonexisting@dummy.it')
        self.assertEqual(resp['message'], u'No user found in db for nonexisting@dummy.it mail address')

        resp = self.proxy.get_user_by_email('nonexisting@dummy.it')
        self.assertEqual(resp['message'], u'No user found in db for nonexisting@dummy.it mail address')

        resp = self.proxy.get_user_by_openid('nonexisting@dummy.it')
        self.assertEqual(resp['message'], u'No user found in db for nonexisting@dummy.it openid')

    def test_not_found_customer(self):
        resp = self.proxy.get_customer_by_name('nonexistingcustomer')
        self.assertEqual(resp['message'], u'No customer found in db for nonexistingcustomer name')

    def test_not_found_project(self):
        resp = self.proxy.get_project_by_name('nonexistingproject')
        self.assertEqual(resp['message'], u'No project found in db for nonexistingproject name')

    def test_time_entry_creation(self):
        """
        This test check time entry parameters
        """
        #customer data
        customer_name = u'RFCCustomer'
        #project data
        project_name = u'A new project'
        project_id = 'a-new-project'
        #entry data
        entry_date = datetime.date(2011, 05, 26)
        entry_hours = '2:30'
        entry_location = u'RedTurtle Technology'
        entry_description = u'Trying to create ticket for API tests'
        entry_ticket = '45'

        #Start to create customer, project and time entry for project
        session = DBSession()
        project = Project(name=project_name, id=project_id)
        customer = Customer(name=customer_name)
        customer.add_project(project)
        session.add(customer)
        transaction.commit()

        #Try to get errors
        resp = self.proxy.create_new_simple_time_entry(1, entry_date,
                                                       entry_hours, entry_description,
                                                       entry_location, project_id)
        self.assertEqual(resp['message'], u"'int' object has no attribute 'decode'")

        resp = self.proxy.create_new_simple_time_entry(entry_ticket,
                                                       entry_date,
                                                       u'9000',
                                                       entry_description,
                                                       entry_location,
                                                       project_id)

        self.assertEqual(resp['message'], u'Cannot parse time (must be HH:MM)')

        resp = self.proxy.create_new_simple_time_entry(entry_ticket,
                                                       entry_date,
                                                       u'19:40',
                                                       entry_description,
                                                       entry_location,
                                                       project_id)

        self.assertEqual(resp['message'], u'Time value too big (must be <= 16:00)')

        resp = self.proxy.create_new_simple_time_entry(entry_ticket,
                                                       entry_date,
                                                       entry_hours,
                                                       entry_description,
                                                       entry_location,
                                                       100)
        self.assertEqual(resp['message'], u'Not able to get the project with id 100')

        resp = self.proxy.create_new_simple_time_entry(entry_ticket,
                                                       '2011 01 01',
                                                       entry_hours,
                                                       entry_description,
                                                       entry_location,
                                                       100)
        self.assertEqual(resp['message'],  u"time data '2011 01 01' does not match format '%Y-%m-%d'")

        #Let's try to create a simple time entry
        resp = self.proxy.create_new_simple_time_entry(entry_ticket,
                                                       entry_date,
                                                       entry_hours,
                                                       entry_description,
                                                       entry_location,
                                                       project_id)

        self.assertRegexpMatches(resp['message'], u'Correctly added time entry \d+ for %s ticket #%s' %(project_id, entry_ticket))

        resp = self.proxy.create_new_simple_time_entry(entry_ticket,
                                                       entry_date,
                                                       entry_hours,
                                                       '',
                                                       entry_location,
                                                       project_id)
        self.assertEqual(resp['message'], u"Description is required.")

        #Now try to create a more complex time entry
        entry_start = datetime.datetime(2011, 01, 01, 15, 30)
        entry_end = datetime.datetime(2011, 01, 01, 17, 30)
        entry_ticket = '#99'

        resp = self.proxy.create_new_advanced_time_entry(99,
                                                         entry_start,
                                                         entry_end,
                                                         entry_description,
                                                         entry_location,
                                                         10)
        self.assertEqual(resp['message'], u"'int' object has no attribute 'decode'")

        resp = self.proxy.create_new_advanced_time_entry(entry_ticket,
                                                         entry_start,
                                                         entry_end,
                                                         entry_description,
                                                         entry_location,
                                                         100)
        self.assertEqual(resp['message'], u'Not able to get the project with id 100')

        resp = self.proxy.create_new_advanced_time_entry(entry_ticket,
                                                         '2011 08 24',
                                                         entry_end,
                                                         entry_description,
                                                         entry_location,
                                                         10)
        self.assertEqual(resp['message'], u"time data '2011 08 24' does not match format '%Y-%m-%d %H:%M:%S'")

        resp = self.proxy.create_new_advanced_time_entry(entry_ticket,
                                                         entry_start,
                                                         entry_end,
                                                         entry_description,
                                                         entry_location,
                                                         project_id)
        self.assertRegexpMatches(resp['message'], u'Correctly added time entry \d+ for %s ticket #%s' %(project_id, entry_ticket))


if __name__ == '__main__':
    unittest2.main()



# -*- coding: utf-8 -*-

"""
Common fields for all the reports, to be cloned before use.
"""

import colander
import deform

from colander import SchemaNode
from deform_bootstrap.widget import ChosenSingleWidget, ChosenMultipleWidget


customer_id = SchemaNode(typ=colander.String(),
                         widget=ChosenSingleWidget(css_class='customer-select',
                                                   placeholder=u'Select customer'),
                         missing=colander.null,
                         title=u'')

project_id =SchemaNode(typ=colander.String(),
                       widget=ChosenSingleWidget(css_class='project-select',
                                                 placeholder=u'Select project'),
                       missing=colander.null,
                       title=u'')

customer_requests = SchemaNode(deform.Set(allow_empty=True),
                               widget=ChosenMultipleWidget(css_class='customer-request-select',
                                                           placeholder=u'Select customer requests'),
                               missing=colander.null,
                               title=u'')

contracts = SchemaNode(deform.Set(allow_empty=True),
                       widget=ChosenMultipleWidget(css_class='contract-select',
                                                   placeholder=u'Select contract'),
                       missing=colander.null,
                       title=u'')

date_from = SchemaNode(typ=colander.Date(),
                       widget=deform.widget.DateInputWidget(size=11,
                                                            placeholder=u'from'),
                       missing=colander.null,
                       title=u'')

date_to = SchemaNode(typ=colander.Date(),
                     widget=deform.widget.DateInputWidget(size=11,
                                                          placeholder=u'to'),
                     missing=colander.null,
                     title=u'')

users = SchemaNode(deform.Set(allow_empty=True),
                   widget=ChosenMultipleWidget(placeholder=u'Select people'),
                   missing=colander.null,
                   title=u'')


workflow_states = SchemaNode(deform.Set(allow_empty=True),
                             widget=ChosenMultipleWidget(placeholder=u'State'),
                             missing=colander.null,
                             title=u'')

invoice_number = SchemaNode(typ=colander.String(),
                            missing=colander.null,
                            widget = deform.widget.TextInputWidget(placeholder=u'invoice nr.'),
                            title=u'')

searchtext = SchemaNode(colander.String(),
                        widget = deform.widget.TextInputWidget(css_class='xlarge',
                                                               placeholder=u'description'),
                        missing = colander.null,
                        title = u'')



import unittest2
import transaction

from pyramid import testing
from pyramid_mailer.mailer import DummyMailer
from pyramid_mailer.interfaces import IMailer

from penelope.core import messages
from penelope.core.models.dashboard import User, PasswordResetToken
from penelope.core.models import DBSession


class Request(testing.DummyRequest):
    """just for testing"""

    def __init__(self, **kwargs):
        self.msgs = messages.Messages()
        testing.DummyRequest.__init__(self, **kwargs)
        mailer = DummyMailer()
        self.registry.registerUtility(mailer, IMailer)

    def add_message(self, *args, **kwargs):
        self.msgs.add(*args, **kwargs)


class TestSecurity(unittest2.TestCase):

    def test_password_reset_for_no_user(self):
        from penelope.core.security.views import password_reset 
        request = Request(params={'email':u'notexistinguser'})
        self.assertFalse(password_reset(request)['token'])

    def add_user(self, email):
        session = DBSession()
        user = User(email=email)
        session.add(user)
        transaction.commit()

    def generate_token(self, email):
        from penelope.core.security.views import password_reset 
        request = Request(params={'email':email})
        return password_reset(request)

    def test_autoregister(self):
        from penelope.core.security.views import autoregister

        profile = {'emails': [u'user0@dummy.it'],
                   'displayName': u'user0',
                   'accounts': [{'username': u'user0'}]}
        user = autoregister(profile)
        self.assertEqual(user.email, 'user0@dummy.it')

    def test_password_reset_for_existing_user(self):
        email = u'user1@dummy.it'
        self.add_user(email)
        self.assertTrue(self.generate_token(email)['token'])

    def test_token_store(self):
        email = u'user2@dummy.it'
        self.add_user(email)
        token = self.generate_token(email)['token']
        session = DBSession()
        self.assertEqual(session.query(PasswordResetToken).filter_by(token=token).one().user.email, email)

    def test_token_password_change(self):
        from pyramid_mailer import get_mailer
        email = u'user3@dummy.it'
        self.add_user(email)
        response = self.generate_token(email)
        mailer = get_mailer(response['request'])
        self.assertTrue(len(mailer.outbox))

        from penelope.core.security.views import change_password
        request = Request(method='POST', params={'token': response['token'],
                                                 'password': 'topsecret',
                                                 'password_repeat': 'topsecret'})
        response = change_password(request)
        self.assertEqual(response.headers.get('Location'),'/login_form')

    def test_token_password_change_wrong_token(self):
        from penelope.core.security.views import change_password
        request = Request(method='POST', params={'token': u'notexistingtoken',
                                                 'password': 'topsecret',
                                                 'password_repeat': 'topsecret'})
        response = change_password(request)
        self.assertEqual(response['request'].msgs[0].message, u'Token doesn\'t exist.')

    def test_token_password_change_missmatch(self):
        email = u'user5@dummy.it'
        self.add_user(email)
        token = self.generate_token(email)['token']
        from penelope.core.security.views import change_password
        request = Request(method='POST', params={'token': token,
                                                 'password': 'thisisverylong',
                                                 'password_repeat': 'missmatch'})
        response = change_password(request)
        self.assertEqual(response['request'].msgs[0].message, u'Passwords missmatch.')


    def test_token_store_cleanup(self):
        email = u'user6@dummy.it'
        self.add_user(email)
        self.generate_token(email)['token']
        token2 = self.generate_token(email)['token']
        session = DBSession()
        self.assertEqual(len(session.query(PasswordResetToken).join(User).filter(User.email == email).all()),1)
        from penelope.core.security.views import change_password
        request = Request(method='POST', params={'token': token2,
                                                 'password': 'topsecret',
                                                 'password_repeat': 'topsecret'})
        response = change_password(request)
        self.assertEqual(response.headers.get('Location'),'/login_form')
        self.assertEqual(len(session.query(PasswordResetToken).join(User).filter(User.email == email).all()),0)

    def test_password_set(self):
        email = u'user7@dummy.it'
        self.add_user(email)
        session = DBSession()
        self.assertEqual(session.query(User).filter_by(email=email).one().password, None)
        token = self.generate_token(email)['token']
        from penelope.core.security.views import change_password
        request = Request(method='POST', params={'token': token,
                                                 'password': 'topsecret',
                                                 'password_repeat': 'topsecret'})
        response = change_password(request)
        self.assertEqual(response.headers.get('Location'),'/login_form')
        self.assertNotEqual(session.query(User).filter_by(email=email).one().password, None)



if __name__ == '__main__':
    unittest2.main()



import zope.interface

from repoze.who.interfaces import IChallengeDecider
from repoze.who.plugins.friendlyform import FriendlyFormPlugin as BasePlugin


def challenge_decider(environ, status, headers):
    return status.startswith('403') or status.startswith('401')
zope.interface.directlyProvides(challenge_decider, IChallengeDecider)


class FriendlyFormPlugin(BasePlugin):
    def _get_full_path(self, path, environ):
        """
        Return the full path to ``path`` by prepending the SCRIPT_NAME.
        If ``path`` is a URL, do nothing.
        BBB: If ``path`` start with double /, do nothing
        """
        if path.startswith('//'):
            path = path[1:]
        elif path.startswith('/'):
            path = environ.get('SCRIPT_NAME', '') + path
        return path


def make_plugin(login_form_url, login_handler_path, logout_handler_path,
                rememberer_name, post_login_url=None, post_logout_url=None,
                login_counter_name=None):

    if login_form_url is None:
        raise ValueError(
            'must include login_form_url in configuration')
    if login_handler_path is None:
        raise ValueError(
            'login_handler_path must not be None')
    if logout_handler_path is None:
        raise ValueError(
            'logout_handler_path must not be None')
    if rememberer_name is None:
        raise ValueError(
            'must include rememberer key (name of another IIdentifier plugin)')

    plugin = FriendlyFormPlugin(login_form_url,
                                login_handler_path,
                                post_login_url,
                                logout_handler_path,
                                post_logout_url,
                                rememberer_name,
                                login_counter_name,
                                charset=None)
    return plugin


def rolefinder(identity, request):
    roles = set()
    user = request.authenticated_user or identity.get('user')
    if user and user.active:
        context = request.challenge_item or request.model_instance
        roles = set(['role:%s' % a for a in user.roles_in_context(context)])
    return roles



#import json
import string

import mandrill
from pyramid.httpexceptions import HTTPForbidden
from pyramid.httpexceptions import HTTPFound
from pyramid.i18n import TranslationStringFactory
from pyramid.security import authenticated_userid
from pyramid.security import remember, forget
from pyramid.view import view_config
#from pyramid_mailer import get_mailer
#from pyramid_mailer.message import Message
from pyramid_skins import SkinObject
from random import choice
from sqlalchemy.orm.exc import NoResultFound
from uuid import uuid4

from penelope.core.models import DBSession
from penelope.core.models.dashboard import User, PasswordResetToken

_ = TranslationStringFactory('penelope')


def autoregister(profile):
    session = DBSession()
    chars = string.letters + string.digits
    password = ''.join(choice(chars) for _ in range(9))
    user = User(fullname = profile.get('displayName'),
                email = profile.get('emails')[0])
    user.set_password(password)
    user.add_openid(profile.get('accounts')[0].get('username'))
    session.add(user)
    return user


def forbidden(context, request, *challenge_args):
    if authenticated_userid(request):
        result =  {'context':context,
                   'request':request}
        return result
    else:
        return HTTPForbidden()


@view_config(context='velruse.api.AuthenticationComplete')
def login(context, request):
    headers={}
    emails = context.profile.get('emails')
    if emails:
        try:
            user = DBSession.query(User).filter(User.email.in_(emails)).one()
        except NoResultFound: #we are in a proper domain - we don't have a user
            user = autoregister(context.profile)
        if not user.active:
            raise HTTPForbidden()
        headers = remember(request, str(user.login))
    return login_success(context, request, headers=headers)


@view_config(name='login_success',permission='view_anon')
def login_success(context, request, headers=None):
    location= request.session.get('came_from')
    if not location or 'favicon' in location:
        location = '/'
    return HTTPFound(location=location, headers=headers)


@view_config(name='logout', permission='view_anon')
def logout(request):
    headers = forget(request)
    request.add_message(_(u'You have been loggout.'),'success')
    return HTTPFound(location='/', headers=headers)


@view_config(name='login_form', renderer='skin', permission='view_anon')
def login_form(context, request):
    """ Return login_form """
    came_from = '/'
    request.session['came_from'] = came_from
    if request.params.get('came_from') and not 'log' in request.params.get('came_from'):
        came_from = request.params.get('came_from')
        request.session['came_from'] = came_from 
    return {'context':context,
            'came_from': came_from,
            'request':request}


@view_config(name='password_reset_form', renderer='skin', permission='view_anon')
def password_reset_form(request):
    return {'request': request}


@view_config(name='password_reset', request_method='POST', renderer='skin', permission='view_anon')
def password_reset(request):
    email = request.params.get('email')
    try:
        session = DBSession()
        user = DBSession.query(User).filter_by(email=email).one()
        ptoken = DBSession.query(PasswordResetToken).get(user.id)
        if not ptoken:
            ptoken = PasswordResetToken(user_id=user.id)
        token = str(uuid4())
        ptoken.token = token 
        session.add(ptoken)
    except NoResultFound:
        token = None

    if token:
        settings = request.registry.settings
        mandrill_client = mandrill.Mandrill(settings['mail.password'])

        from_addr = settings['mail.default_sender']
        params = {"header": u'Password reset',
                  "message": u'Please click on the link bellow to reset your penelope account\'s password.',
                  "link": '%s/change_password?token=%s' % (request.application_url, token),
                  "action": 'Reset password'}
        merged_params = []
        for k,v in params.items():
            merged_params.append({'name': k, 'content':v})

        message = {'auto_html': None,
                   'auto_text': None,
                   'from_email': from_addr,
                   'from_name': 'RedTurtle Team',
                   'headers': {'Reply-To': from_addr},
                   'important': True,
                   'inline_css': True,
                   'global_merge_vars': merged_params,
                   'subject': u"Password reset request",
                   'to': [{'email': email}],
                   }
        mandrill_client.messages.send_template(template_name='general',
                                               template_content=[],
                                               message=message)

    return {'request': request, 'token': token}


@view_config(name='change_password', permission='view_anon', request_method='GET', request_param='token')
def change_password_form(context, request):
    token = request.params.get('token')
    return SkinObject('change_password')(request=request, token=token, context=context)


@view_config(name='change_password', renderer='skin', permission='view_anon', request_method='POST', request_param='token')
def change_password(request):
    session = DBSession()
    token = request.params.get('token')
    ptoken = session.query(PasswordResetToken).filter_by(token=token).first()
    password = request.params.get('password')
    repeat = request.params.get('password_repeat')

    if not ptoken:
        request.add_message(_(u'Token doesn\'t exist.'), type='danger')
    elif password != repeat:
        request.add_message(_(u'Passwords missmatch.'), type='danger')
    elif not password or not repeat:
        request.add_message(_(u'Missing password.'), type='danger')
    elif len(password)<6:
        request.add_message(_(u'Password too short. It needs to be at least 6 characters long.'), type='danger')
    else:
        ptoken.user.set_password(password)
        session.delete(ptoken)
        request.add_message(_(u'Password changed.'), type='success')
        return HTTPFound(location='/login_form')
    return {'request': request, 'token': token}
